import os
from typing import Any, Dict, Mapping

import pytest

from gql import Client, FileVar, gql
from gql.transport.exceptions import (
    TransportAlreadyConnected,
    TransportClosed,
    TransportConnectionFailed,
    TransportProtocolError,
    TransportQueryError,
    TransportServerError,
)

from .conftest import (
    TemporaryFile,
    get_localhost_ssl_context_client,
    make_upload_handler,
)

# Marking all tests in this file with the httpx marker
pytestmark = pytest.mark.httpx

query1_str = """
    query getContinents {
      continents {
        code
        name
      }
    }
"""

query1_server_answer = (
    '{"data":{"continents":['
    '{"code":"AF","name":"Africa"},{"code":"AN","name":"Antarctica"},'
    '{"code":"AS","name":"Asia"},{"code":"EU","name":"Europe"},'
    '{"code":"NA","name":"North America"},{"code":"OC","name":"Oceania"},'
    '{"code":"SA","name":"South America"}]}}'
)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_httpx_query(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    async def handler(request):
        return web.Response(
            text=query1_server_answer,
            content_type="application/json",
            headers={"dummy": "test1234"},
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    def test_code():
        transport = HTTPXTransport(url=url)

        with Client(transport=transport) as session:

            query = gql(query1_str)

            # Execute query synchronously
            result = session.execute(query)

            continents = result["continents"]

            africa = continents[0]

            assert africa["code"] == "AF"

            # Checking response headers are saved in the transport
            assert hasattr(transport, "response_headers")
            assert isinstance(transport.response_headers, Mapping)
            assert transport.response_headers["dummy"] == "test1234"

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
@pytest.mark.parametrize("verify_https", ["disabled", "cert_provided"])
async def test_httpx_query_https(ssl_aiohttp_server, run_sync_test, verify_https):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    async def handler(request):
        return web.Response(
            text=query1_server_answer,
            content_type="application/json",
            headers={"dummy": "test1234"},
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await ssl_aiohttp_server(app)

    url = str(server.make_url("/"))

    assert str(url).startswith("https://")

    def test_code():
        extra_args = {}

        if verify_https == "cert_provided":
            _, ssl_context = get_localhost_ssl_context_client()

            extra_args["verify"] = ssl_context
        elif verify_https == "disabled":
            extra_args["verify"] = False

        transport = HTTPXTransport(
            url=url,
            **extra_args,
        )

        with Client(transport=transport) as session:

            query = gql(query1_str)

            # Execute query synchronously
            result = session.execute(query)

            continents = result["continents"]

            africa = continents[0]

            assert africa["code"] == "AF"

            # Checking response headers are saved in the transport
            assert hasattr(transport, "response_headers")
            assert isinstance(transport.response_headers, Mapping)
            assert transport.response_headers["dummy"] == "test1234"

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
@pytest.mark.parametrize("verify_https", ["explicitely_enabled", "default"])
async def test_httpx_query_https_self_cert_fail(
    ssl_aiohttp_server, run_sync_test, verify_https
):
    """By default, we should verify the ssl certificate"""
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    async def handler(request):
        return web.Response(
            text=query1_server_answer,
            content_type="application/json",
            headers={"dummy": "test1234"},
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await ssl_aiohttp_server(app)

    url = str(server.make_url("/"))

    assert str(url).startswith("https://")

    def test_code():
        extra_args: Dict[str, Any] = {}

        if verify_https == "explicitely_enabled":
            extra_args["verify"] = True

        transport = HTTPXTransport(
            url=url,
            **extra_args,
        )

        query = gql(query1_str)

        expected_error = "certificate verify failed: self-signed certificate"

        with pytest.raises(TransportConnectionFailed) as exc_info:
            with Client(transport=transport) as session:
                session.execute(query)

        assert expected_error in str(exc_info.value)

        with pytest.raises(TransportConnectionFailed) as exc_info:
            with Client(transport=transport) as session:
                session.execute_batch([query])

        assert expected_error in str(exc_info.value)

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_httpx_cookies(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    async def handler(request):
        assert "COOKIE" in request.headers
        assert "cookie1=val1" == request.headers["COOKIE"]

        return web.Response(text=query1_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    def test_code():
        transport = HTTPXTransport(url=url, cookies={"cookie1": "val1"})

        with Client(transport=transport) as session:

            query = gql(query1_str)

            # Execute query synchronously
            result = session.execute(query)

            continents = result["continents"]

            africa = continents[0]

            assert africa["code"] == "AF"

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_httpx_error_code_401(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    async def handler(request):
        # Will generate http error code 401
        return web.Response(
            text='{"error":"Unauthorized","message":"401 Client Error: Unauthorized"}',
            content_type="application/json",
            status=401,
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    def test_code():
        transport = HTTPXTransport(url=url)

        with Client(transport=transport) as session:

            query = gql(query1_str)

            with pytest.raises(TransportServerError) as exc_info:
                session.execute(query)

            assert "Client error '401 Unauthorized'" in str(exc_info.value)

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_httpx_error_code_429(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    async def handler(request):
        # Will generate http error code 429
        return web.Response(
            text="""
<html>
  <head>
     <title>Too Many Requests</title>
  </head>
  <body>
     <h1>Too Many Requests</h1>
     <p>I only allow 50 requests per hour to this Web site per
        logged in user.  Try again soon.</p>
  </body>
</html>""",
            content_type="text/html",
            status=429,
            headers={"Retry-After": "3600"},
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    def test_code():
        transport = HTTPXTransport(url=url)

        with Client(transport=transport) as session:

            query = gql(query1_str)

            with pytest.raises(TransportServerError) as exc_info:
                session.execute(query)

            assert "429, message='Too Many Requests'" in str(exc_info.value)

        # Checking response headers are saved in the transport
        assert hasattr(transport, "response_headers")
        assert isinstance(transport.response_headers, Mapping)
        assert transport.response_headers["Retry-After"] == "3600"


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_httpx_error_code_500(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    async def handler(request):
        # Will generate http error code 500
        raise Exception("Server error")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    def test_code():
        transport = HTTPXTransport(url=url)

        with Client(transport=transport) as session:

            query = gql(query1_str)

            with pytest.raises(TransportServerError):
                session.execute(query)

    await run_sync_test(server, test_code)


query1_server_error_answer = '{"errors": ["Error 1", "Error 2"]}'


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_httpx_error_code(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    async def handler(request):
        return web.Response(
            text=query1_server_error_answer, content_type="application/json"
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    def test_code():
        transport = HTTPXTransport(url=url)

        with Client(transport=transport) as session:

            query = gql(query1_str)

            with pytest.raises(TransportQueryError):
                session.execute(query)

    await run_sync_test(server, test_code)


invalid_protocol_responses = [
    "{}",
    "qlsjfqsdlkj",
    '{"not_data_or_errors": 35}',
    "",
]


@pytest.mark.aiohttp
@pytest.mark.asyncio
@pytest.mark.parametrize("response", invalid_protocol_responses)
async def test_httpx_invalid_protocol(aiohttp_server, response, run_sync_test):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    async def handler(request):
        return web.Response(text=response, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    def test_code():
        transport = HTTPXTransport(url=url)

        with Client(transport=transport) as session:

            query = gql(query1_str)

            with pytest.raises(TransportProtocolError):
                session.execute(query)

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_httpx_cannot_connect_twice(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    async def handler(request):
        return web.Response(text=query1_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    def test_code():
        transport = HTTPXTransport(url=url)

        with Client(transport=transport) as session:

            with pytest.raises(TransportAlreadyConnected):
                session.transport.connect()

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_httpx_cannot_execute_if_not_connected(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    async def handler(request):
        return web.Response(text=query1_server_answer, content_type="application/json")

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    def test_code():
        transport = HTTPXTransport(url=url)

        query = gql(query1_str)

        with pytest.raises(TransportClosed):
            transport.execute(query)

    await run_sync_test(server, test_code)


query1_server_answer_with_extensions = (
    '{"data":{"continents":['
    '{"code":"AF","name":"Africa"},{"code":"AN","name":"Antarctica"},'
    '{"code":"AS","name":"Asia"},{"code":"EU","name":"Europe"},'
    '{"code":"NA","name":"North America"},{"code":"OC","name":"Oceania"},'
    '{"code":"SA","name":"South America"}]},'
    '"extensions": {"key1": "val1"}'
    "}"
)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_httpx_query_with_extensions(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    async def handler(request):
        return web.Response(
            text=query1_server_answer_with_extensions, content_type="application/json"
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    def test_code():
        transport = HTTPXTransport(url=url)

        with Client(transport=transport) as session:

            query = gql(query1_str)

            execution_result = session.execute(query, get_execution_result=True)

            assert execution_result.extensions["key1"] == "val1"

    await run_sync_test(server, test_code)


file_upload_mutation_1 = """
    mutation($file: Upload!) {
      uploadFile(input:{other_var:$other_var, file:$file}) {
        success
      }
    }
"""

file_upload_mutation_1_operations = (
    '{"query": "mutation ($file: Upload!) {\\n  uploadFile(input: {other_var: '
    '$other_var, file: $file}) {\\n    success\\n  }\\n}", "variables": '
    '{"file": null, "other_var": 42}}'
)

file_upload_mutation_1_map = '{"0": ["variables.file"]}'

file_1_content = """
This is a test file
This file will be sent in the GraphQL mutation
"""


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_httpx_file_upload(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    app = web.Application()
    app.router.add_route(
        "POST",
        "/",
        make_upload_handler(
            expected_map=file_upload_mutation_1_map,
            expected_operations=file_upload_mutation_1_operations,
            expected_contents=[file_1_content],
        ),
    )
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    def test_code():
        transport = HTTPXTransport(url=url)

        with TemporaryFile(file_1_content) as test_file:
            with Client(transport=transport) as session:
                query = gql(file_upload_mutation_1)

                file_path = test_file.filename

                # Using an opened file
                with open(file_path, "rb") as f:

                    query.variable_values = {"file": f, "other_var": 42}
                    with pytest.warns(
                        DeprecationWarning,
                        match="Not using FileVar for file upload is deprecated",
                    ):
                        execution_result = session.execute(query, upload_files=True)

                    assert execution_result["success"]

                # Using an opened file inside a FileVar object
                with open(file_path, "rb") as f:

                    query.variable_values = {"file": FileVar(f), "other_var": 42}
                    execution_result = session.execute(query, upload_files=True)

                    assert execution_result["success"]

                # Using an filename string inside a FileVar object
                query.variable_values = {
                    "file": FileVar(file_path),
                    "other_var": 42,
                }
                execution_result = session.execute(query, upload_files=True)

                assert execution_result["success"]

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_httpx_file_upload_with_content_type(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    app = web.Application()
    app.router.add_route(
        "POST",
        "/",
        make_upload_handler(
            file_headers=[{"Content-Type": "application/pdf"}],
            expected_map=file_upload_mutation_1_map,
            expected_operations=file_upload_mutation_1_operations,
            expected_contents=[file_1_content],
        ),
    )
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    def test_code():
        transport = HTTPXTransport(url=url)

        with TemporaryFile(file_1_content) as test_file:
            with Client(transport=transport) as session:
                query = gql(file_upload_mutation_1)

                file_path = test_file.filename

                # Using an opened file
                with open(file_path, "rb") as f:

                    # Setting the content_type
                    f.content_type = "application/pdf"  # type: ignore

                    query.variable_values = {"file": f, "other_var": 42}
                    with pytest.warns(
                        DeprecationWarning,
                        match="Not using FileVar for file upload is deprecated",
                    ):
                        execution_result = session.execute(query, upload_files=True)

                    assert execution_result["success"]

                # Using FileVar
                query.variable_values = {
                    "file": FileVar(file_path, content_type="application/pdf"),
                    "other_var": 42,
                }
                execution_result = session.execute(query, upload_files=True)

                assert execution_result["success"]

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_httpx_file_upload_default_filename_is_basename(
    aiohttp_server, run_sync_test
):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    app = web.Application()

    with TemporaryFile(file_1_content) as test_file:
        file_path = test_file.filename
        file_basename = os.path.basename(file_path)

        app.router.add_route(
            "POST",
            "/",
            make_upload_handler(
                filenames=[file_basename],
                expected_map=file_upload_mutation_1_map,
                expected_operations=file_upload_mutation_1_operations,
                expected_contents=[file_1_content],
            ),
        )
        server = await aiohttp_server(app)

        url = str(server.make_url("/"))

        def test_code():
            transport = HTTPXTransport(url=url)

            with Client(transport=transport) as session:
                query = gql(file_upload_mutation_1)

                # Using FileVar
                query.variable_values = {
                    "file": FileVar(file_path),
                    "other_var": 42,
                }
                execution_result = session.execute(query, upload_files=True)

                assert execution_result["success"]

        await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_httpx_file_upload_additional_headers(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    app = web.Application()
    app.router.add_route(
        "POST",
        "/",
        make_upload_handler(
            request_headers={"X-Auth": "foobar"},
            expected_map=file_upload_mutation_1_map,
            expected_operations=file_upload_mutation_1_operations,
            expected_contents=[file_1_content],
        ),
    )
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    def test_code():
        transport = HTTPXTransport(url=url, headers={"X-Auth": "foobar"})

        with TemporaryFile(file_1_content) as test_file:
            with Client(transport=transport) as session:
                query = gql(file_upload_mutation_1)

                file_path = test_file.filename

                query.variable_values = {"file": FileVar(file_path), "other_var": 42}
                execution_result = session.execute(query, upload_files=True)

                assert execution_result["success"]

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_httpx_binary_file_upload(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    # This is a sample binary file content containing all possible byte values
    binary_file_content = bytes(range(0, 256))

    app = web.Application()
    app.router.add_route(
        "POST",
        "/",
        make_upload_handler(
            binary=True,
            expected_contents=[binary_file_content],
            expected_map=file_upload_mutation_1_map,
            expected_operations=file_upload_mutation_1_operations,
        ),
    )
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    transport = HTTPXTransport(url=url)

    def test_code():
        with TemporaryFile(binary_file_content) as test_file:
            with Client(transport=transport) as session:

                query = gql(file_upload_mutation_1)

                file_path = test_file.filename

                query.variable_values = {"file": FileVar(file_path), "other_var": 42}

                execution_result = session.execute(query, upload_files=True)

                assert execution_result["success"]

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_httpx_file_upload_two_files(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    file_upload_mutation_2 = """
    mutation($file1: Upload!, $file2: Upload!) {
      uploadFile(input:{file1:$file, file2:$file}) {
        success
      }
    }
    """

    file_upload_mutation_2_operations = (
        '{"query": "mutation ($file1: Upload!, $file2: Upload!) {\\n  '
        'uploadFile(input: {file1: $file, file2: $file}) {\\n    success\\n  }\\n}", '
        '"variables": {"file1": null, "file2": null}}'
    )

    file_upload_mutation_2_map = '{"0": ["variables.file1"], "1": ["variables.file2"]}'

    file_2_content = """
    This is a second test file
    This file will also be sent in the GraphQL mutation
    """

    app = web.Application()
    app.router.add_route(
        "POST",
        "/",
        make_upload_handler(
            nb_files=2,
            expected_map=file_upload_mutation_2_map,
            expected_operations=file_upload_mutation_2_operations,
            expected_contents=[file_1_content, file_2_content],
        ),
    )
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    def test_code():
        transport = HTTPXTransport(url=url)

        with TemporaryFile(file_1_content) as test_file_1:
            with TemporaryFile(file_2_content) as test_file_2:

                with Client(transport=transport) as session:

                    query = gql(file_upload_mutation_2)

                    file_path_1 = test_file_1.filename
                    file_path_2 = test_file_2.filename

                    query.variable_values = {
                        "file1": FileVar(file_path_1),
                        "file2": FileVar(file_path_2),
                    }

                    execution_result = session.execute(query, upload_files=True)

                    assert execution_result["success"]

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_httpx_file_upload_list_of_two_files(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    file_upload_mutation_3 = """
    mutation($files: [Upload!]!) {
      uploadFiles(input:{files:$files}) {
        success
      }
    }
    """

    file_upload_mutation_3_operations = (
        '{"query": "mutation ($files: [Upload!]!) {\\n  uploadFiles'
        "(input: {files: $files})"
        ' {\\n    success\\n  }\\n}", "variables": {"files": [null, null]}}'
    )

    file_upload_mutation_3_map = (
        '{"0": ["variables.files.0"], "1": ["variables.files.1"]}'
    )

    file_2_content = """
    This is a second test file
    This file will also be sent in the GraphQL mutation
    """

    app = web.Application()
    app.router.add_route(
        "POST",
        "/",
        make_upload_handler(
            nb_files=2,
            expected_map=file_upload_mutation_3_map,
            expected_operations=file_upload_mutation_3_operations,
            expected_contents=[file_1_content, file_2_content],
        ),
    )
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    def test_code():
        transport = HTTPXTransport(url=url)

        with TemporaryFile(file_1_content) as test_file_1:
            with TemporaryFile(file_2_content) as test_file_2:
                with Client(transport=transport) as session:

                    query = gql(file_upload_mutation_3)

                    file_path_1 = test_file_1.filename
                    file_path_2 = test_file_2.filename

                    query.variable_values = {
                        "files": [
                            FileVar(file_path_1),
                            FileVar(file_path_2),
                        ],
                    }

                    execution_result = session.execute(query, upload_files=True)

                    assert execution_result["success"]

    await run_sync_test(server, test_code)


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_httpx_error_fetching_schema(aiohttp_server, run_sync_test):
    from aiohttp import web

    from gql.transport.httpx import HTTPXTransport

    error_answer = """
{
    "errors": [
        {
            "errorType": "UnauthorizedException",
            "message": "Permission denied"
        }
    ]
}
"""

    async def handler(request):
        return web.Response(
            text=error_answer,
            content_type="application/json",
        )

    app = web.Application()
    app.router.add_route("POST", "/", handler)
    server = await aiohttp_server(app)

    url = str(server.make_url("/"))

    def test_code():
        transport = HTTPXTransport(url=url)

        with pytest.raises(TransportQueryError) as exc_info:
            with Client(transport=transport, fetch_schema_from_transport=True):
                pass

        expected_error = (
            "Error while fetching schema: "
            "{'errorType': 'UnauthorizedException', 'message': 'Permission denied'}"
        )

        assert expected_error in str(exc_info.value)
        assert transport.client is None

    await run_sync_test(server, test_code)



import abc
from typing import Any, Dict, List, Optional


class AdapterConnection(abc.ABC):
    """Abstract interface for subscription connections.

    This allows different WebSocket implementations to be used interchangeably.
    """

    url: str
    connect_args: Dict[str, Any]
    subprotocols: Optional[List[str]]

    def __init__(self, url: str, connect_args: Optional[Dict[str, Any]]):
        """Initialize the connection adapter."""
        self.url: str = url

        if connect_args is None:
            connect_args = {}
        self.connect_args = connect_args

        self.subprotocols = None

    @abc.abstractmethod
    async def connect(self) -> None:
        """Connect to the server."""
        pass  # pragma: no cover

    @abc.abstractmethod
    async def send(self, message: str) -> None:
        """Send message to the server.

        Args:
            message: String message to send

        Raises:
            TransportConnectionFailed: If connection closed
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    async def receive(self) -> str:
        """Receive message from the server.

        Returns:
            String message received

        Raises:
            TransportConnectionFailed: If connection closed
            TransportProtocolError: If protocol error or binary data received
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    async def close(self) -> None:
        """Close the connection."""
        pass  # pragma: no cover

    @property
    @abc.abstractmethod
    def response_headers(self) -> Dict[str, str]:
        """Get the response headers from the connection.

        Returns:
            Dictionary of response headers
        """
        pass  # pragma: no cover



import asyncio
import json
import types
from typing import List

import pytest

from gql import Client, gql
from gql.transport.exceptions import (
    TransportAlreadyConnected,
    TransportConnectionFailed,
    TransportProtocolError,
    TransportQueryError,
)

from .conftest import MS, WebSocketServerHelper

# Marking all tests in this file with the websockets marker
pytestmark = pytest.mark.websockets

invalid_query_str = """
    query getContinents {
      continents {
        code
        bloh
      }
    }
"""

invalid_query1_server_answer = (
    '{{"type":"data","id":"{query_id}",'
    '"payload":{{"errors":['
    '{{"message":"Cannot query field \\"bloh\\" on type \\"Continent\\".",'
    '"locations":[{{"line":4,"column":5}}],'
    '"extensions":{{"code":"INTERNAL_SERVER_ERROR"}}}}]}}}}'
)

invalid_query1_server = [invalid_query1_server_answer]


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [invalid_query1_server], indirect=True)
@pytest.mark.parametrize("query_str", [invalid_query_str])
async def test_websocket_invalid_query(client_and_server, query_str):

    session, server = client_and_server

    query = gql(query_str)

    with pytest.raises(TransportQueryError) as exc_info:
        await session.execute(query)

    exception = exc_info.value

    assert isinstance(exception.errors, List)

    error = exception.errors[0]

    assert error["extensions"]["code"] == "INTERNAL_SERVER_ERROR"


invalid_subscription_str = """
    subscription getContinents {
      continents {
        code
        bloh
      }
    }
"""


async def server_invalid_subscription(ws):
    await WebSocketServerHelper.send_connection_ack(ws)
    await ws.recv()
    await ws.send(invalid_query1_server_answer.format(query_id=1))
    await WebSocketServerHelper.send_complete(ws, 1)
    await ws.wait_closed()


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_invalid_subscription], indirect=True)
@pytest.mark.parametrize("query_str", [invalid_subscription_str])
async def test_websocket_invalid_subscription(client_and_server, query_str):

    session, server = client_and_server

    query = gql(query_str)

    with pytest.raises(TransportQueryError) as exc_info:
        async for result in session.subscribe(query):
            pass

    exception = exc_info.value

    assert isinstance(exception.errors, List)

    error = exception.errors[0]

    assert error["extensions"]["code"] == "INTERNAL_SERVER_ERROR"


connection_error_server_answer = (
    '{"type":"connection_error","id":null,'
    '"payload":{"message":"Unexpected token Q in JSON at position 0"}}'
)


async def server_no_ack(ws):
    await ws.wait_closed()


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_no_ack], indirect=True)
@pytest.mark.parametrize("query_str", [invalid_query_str])
async def test_websocket_server_does_not_send_ack(server, query_str):
    from gql.transport.websockets import WebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"

    transport = WebsocketsTransport(url=url, ack_timeout=0.1)

    with pytest.raises(asyncio.TimeoutError):
        async with Client(transport=transport):
            pass


async def server_connection_error(ws):
    await WebSocketServerHelper.send_connection_ack(ws)
    result = await ws.recv()
    print(f"Server received: {result}")
    await ws.send(connection_error_server_answer)
    await ws.wait_closed()


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_connection_error], indirect=True)
@pytest.mark.parametrize("query_str", [invalid_query_str])
async def test_websocket_sending_invalid_data(client_and_server, query_str):

    session, server = client_and_server

    invalid_data = "QSDF"
    print(f">>> {invalid_data}")
    await session.transport.adapter.websocket.send(invalid_data)

    await asyncio.sleep(2 * MS)


invalid_payload_server_answer = (
    '{"type":"error","id":"1","payload":{"message":"Must provide document"}}'
)


async def server_invalid_payload(ws):
    await WebSocketServerHelper.send_connection_ack(ws)
    result = await ws.recv()
    print(f"Server received: {result}")
    await ws.send(invalid_payload_server_answer)
    await WebSocketServerHelper.wait_connection_terminate(ws)
    await ws.wait_closed()


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_invalid_payload], indirect=True)
@pytest.mark.parametrize("query_str", [invalid_query_str])
async def test_websocket_sending_invalid_payload(client_and_server, query_str):

    session, server = client_and_server

    # Monkey patching the _send_query method to send an invalid payload

    async def monkey_patch_send_query(
        self,
        document,
        variable_values=None,
        operation_name=None,
    ):
        query_id = self.next_query_id
        self.next_query_id += 1

        query_str = json.dumps(
            {"id": str(query_id), "type": "start", "payload": "BLAHBLAH"}
        )

        await self._send(query_str)
        return query_id

    session.transport._send_query = types.MethodType(
        monkey_patch_send_query, session.transport
    )

    query = gql(query_str)

    with pytest.raises(TransportQueryError) as exc_info:
        await session.execute(query)

    exception = exc_info.value

    assert isinstance(exception.errors, List)

    error = exception.errors[0]

    assert error["message"] == "Must provide document"


not_json_answer = ["BLAHBLAH"]
missing_type_answer = ["{}"]
missing_id_answer_1 = ['{"type": "data"}']
missing_id_answer_2 = ['{"type": "error"}']
missing_id_answer_3 = ['{"type": "complete"}']
data_without_payload = ['{"type": "data", "id":"1"}']
error_without_payload = ['{"type": "error", "id":"1"}']
payload_is_not_a_dict = ['{"type": "data", "id":"1", "payload": "BLAH"}']
empty_payload = ['{"type": "data", "id":"1", "payload": {}}']
sending_bytes = [b"\x01\x02\x03"]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "server",
    [
        not_json_answer,
        missing_type_answer,
        missing_id_answer_1,
        missing_id_answer_2,
        missing_id_answer_3,
        data_without_payload,
        error_without_payload,
        payload_is_not_a_dict,
        empty_payload,
        sending_bytes,
    ],
    indirect=True,
)
async def test_websocket_transport_protocol_errors(client_and_server):

    session, server = client_and_server

    query = gql("query { hello }")

    with pytest.raises(TransportProtocolError):
        await session.execute(query)


async def server_without_ack(ws):
    # Sending something else than an ack
    await WebSocketServerHelper.send_complete(ws, 1)
    await ws.wait_closed()


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_without_ack], indirect=True)
async def test_websocket_server_does_not_ack(server):
    from gql.transport.websockets import WebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    transport = WebsocketsTransport(url=url)

    with pytest.raises(TransportProtocolError):
        async with Client(transport=transport):
            pass


async def server_closing_directly(ws):
    await ws.close()


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_closing_directly], indirect=True)
async def test_websocket_server_closing_directly(server):
    from gql.transport.websockets import WebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    transport = WebsocketsTransport(url=url)

    with pytest.raises(TransportConnectionFailed):
        async with Client(transport=transport):
            pass


async def server_closing_after_ack(ws):
    await WebSocketServerHelper.send_connection_ack(ws)
    await ws.close()


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_closing_after_ack], indirect=True)
async def test_websocket_server_closing_after_ack(client_and_server):

    session, server = client_and_server

    query = gql("query { hello }")

    print("\n Trying to execute first query.\n")

    with pytest.raises(TransportConnectionFailed) as exc1:
        await session.execute(query)

    exc1_cause = exc1.value.__cause__
    exc1_cause_str = f"{type(exc1_cause).__name__}:{exc1_cause!s}"

    print(f"\n First query Exception cause: {exc1_cause_str}\n")

    assert (
        exc1_cause_str == "ConnectionClosedOK:received 1000 (OK); then sent 1000 (OK)"
    )

    await session.transport.wait_closed()

    print("\n Trying to execute second query.\n")

    with pytest.raises(TransportConnectionFailed) as exc2:
        await session.execute(query)

    exc2_cause = exc2.value.__cause__
    exc2_cause_str = f"{type(exc2_cause).__name__}:{exc2_cause!s}"

    print(f" Second query Exception cause: {exc2_cause_str}\n")

    assert (
        exc2_cause_str == "ConnectionClosedOK:received 1000 (OK); then sent 1000 (OK)"
    )


async def server_sending_invalid_query_errors(ws):
    await WebSocketServerHelper.send_connection_ack(ws)
    invalid_error = (
        '{"type":"error","id":"404","payload":'
        '{"message":"error for no good reason on non existing query"}}'
    )
    await ws.send(invalid_error)
    await ws.wait_closed()


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_sending_invalid_query_errors], indirect=True)
async def test_websocket_server_sending_invalid_query_errors(server):
    from gql.transport.websockets import WebsocketsTransport

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    transport = WebsocketsTransport(url=url)

    # Invalid server message is ignored
    async with Client(transport=transport):
        await asyncio.sleep(2 * MS)


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server_sending_invalid_query_errors], indirect=True)
async def test_websocket_non_regression_bug_105(server):
    from gql.transport.websockets import WebsocketsTransport

    # This test will check a fix to a race condition which happens if the user is trying
    # to connect using the same client twice at the same time
    # See bug #105

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    transport = WebsocketsTransport(url=url)

    client = Client(transport=transport)

    # Create a coroutine which start the connection with the transport but does nothing
    async def client_connect(client):
        async with client:
            await asyncio.sleep(2 * MS)

    # Create two tasks which will try to connect using the same client (not allowed)
    connect_task1 = asyncio.ensure_future(client_connect(client))
    connect_task2 = asyncio.ensure_future(client_connect(client))

    with pytest.raises(TransportAlreadyConnected):
        await asyncio.gather(connect_task1, connect_task2)


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [invalid_query1_server], indirect=True)
async def test_websocket_using_cli_invalid_query(server, monkeypatch, capsys):

    url = f"ws://{server.hostname}:{server.port}/graphql"
    print(f"url = {url}")

    import io

    from gql.cli import get_parser, main

    parser = get_parser(with_examples=True)
    args = parser.parse_args([url])

    # Monkeypatching sys.stdin to simulate getting the query
    # via the standard input
    monkeypatch.setattr("sys.stdin", io.StringIO(invalid_query_str))

    # Flush captured output
    captured = capsys.readouterr()

    await main(args)

    # Check that the error has been printed on stdout
    captured = capsys.readouterr()
    captured_err = str(captured.err).strip()
    print(f"Captured: {captured_err}")

    expected_error = 'Cannot query field "bloh" on type "Continent"'

    assert expected_error in captured_err



import os
from contextlib import suppress
from typing import Any
from unittest import mock

import pytest
from graphql import ExecutionResult, build_ast_schema, parse

from gql import Client, GraphQLRequest, gql
from gql.transport import Transport
from gql.transport.exceptions import TransportQueryError

with suppress(ModuleNotFoundError):
    from urllib3.exceptions import NewConnectionError


@pytest.fixture
def http_transport_query():
    return gql(
        """
        query getContinents {
          continents {
            code
            name
          }
        }
        """
    )


def test_request_transport_not_implemented(http_transport_query):
    class RandomTransport(Transport):
        pass

    with pytest.raises(TypeError) as exc_info:
        RandomTransport()  # type: ignore

    assert "Can't instantiate abstract class RandomTransport" in str(exc_info.value)

    class RandomTransport2(Transport):
        def execute(
            self,
            request: GraphQLRequest,
            *args: Any,
            **kwargs: Any,
        ) -> ExecutionResult:
            return ExecutionResult()

    with pytest.raises(NotImplementedError) as exc_info2:
        RandomTransport2().execute_batch([])

    assert "This Transport has not implemented the execute_batch method" == str(
        exc_info2.value
    )


@pytest.mark.requests
@mock.patch("urllib3.connection.HTTPConnection._new_conn")
def test_retries_on_transport(execute_mock):
    """Testing retries on the transport level

    This forces us to override low-level APIs because the retry mechanism on the urllib3
    (which uses requests) is pretty low-level itself.
    """
    from gql.transport.requests import RequestsHTTPTransport

    expected_retries = 3
    execute_mock.side_effect = NewConnectionError(
        "Should be HTTPConnection", "Fake connection error"  # type: ignore
    )
    transport = RequestsHTTPTransport(
        url="http://127.0.0.1:8000/graphql",
        retries=expected_retries,
    )
    client = Client(transport=transport)

    query = gql(
        """
        {
          myFavoriteFilm: film(id:"RmlsbToz") {
            id
            title
            episodeId
          }
        }
        """
    )
    with client as session:  # We're using the client as context manager
        with pytest.raises(Exception):
            session.execute(query)

    # This might look strange compared to the previous test, but making 3 retries
    # means you're actually doing 4 calls.
    assert execute_mock.call_count == expected_retries + 1

    execute_mock.reset_mock()
    queries = [query, query, query]

    with client as session:  # We're using the client as context manager
        with pytest.raises(Exception):
            session.execute_batch(queries)

    # This might look strange compared to the previous test, but making 3 retries
    # means you're actually doing 4 calls.
    assert execute_mock.call_count == expected_retries + 1


def test_no_schema_no_transport_exception():
    with pytest.raises(AssertionError) as exc_info:
        Client()
    assert "You need to provide either a transport or a schema to the Client." in str(
        exc_info.value
    )


@pytest.mark.online
@pytest.mark.requests
def test_execute_result_error():

    from gql.transport.requests import RequestsHTTPTransport

    client = Client(
        transport=RequestsHTTPTransport(url="https://countries.trevorblades.com/"),
    )

    failing_query = gql(
        """
        query getContinents {
          continents {
            code
            name
            id
          }
        }
        """
    )

    with pytest.raises(TransportQueryError) as exc_info:
        client.execute(failing_query)
    assert 'Cannot query field "id" on type "Continent".' in str(exc_info.value)

    """
    Batching is not supported anymore on countries backend

    with pytest.raises(TransportQueryError) as exc_info:
        client.execute_batch([GraphQLRequest(failing_query)])
    assert 'Cannot query field "id" on type "Continent".' in str(exc_info.value)
    """


@pytest.mark.online
@pytest.mark.requests
def test_http_transport_verify_error(http_transport_query):
    from gql.transport.requests import RequestsHTTPTransport

    with Client(
        transport=RequestsHTTPTransport(
            url="https://countries.trevorblades.com/",
            verify=False,
        )
    ) as client:
        with pytest.warns(Warning) as record:
            client.execute(http_transport_query)

        assert len(record) == 1
        assert "Unverified HTTPS request is being made to host" in str(
            record[0].message
        )

        """
        Batching is not supported anymore on countries backend

        with pytest.warns(Warning) as record:
            client.execute_batch([GraphQLRequest(http_transport_query)])

        assert len(record) == 1
        assert "Unverified HTTPS request is being made to host" in str(
            record[0].message
        )
        """


@pytest.mark.online
@pytest.mark.requests
def test_http_transport_specify_method_valid(http_transport_query):
    from gql.transport.requests import RequestsHTTPTransport

    with Client(
        transport=RequestsHTTPTransport(
            url="https://countries.trevorblades.com/",
            method="POST",
        )
    ) as client:
        result = client.execute(http_transport_query)
        assert result is not None

        """
        Batching is not supported anymore on countries backend

        result = client.execute_batch([GraphQLRequest(http_transport_query)])
        assert result is not None
        """


def test_gql():
    sample_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "fixtures",
        "graphql",
        "sample.graphql",
    )
    with open(sample_path) as source:
        document = parse(source.read())

    schema = build_ast_schema(document)

    query = gql(
        """
        query getUser {
          user(id: "1000") {
            id
            username
          }
        }
        """
    )

    client = Client(schema=schema)
    result = client.execute(query)
    assert result["user"] is None


@pytest.mark.requests
def test_sync_transport_close_on_schema_retrieval_failure():
    """
    Ensure that the transport session is closed if an error occurs when
    entering the context manager (e.g., because schema retrieval fails)
    """

    from gql.transport.requests import RequestsHTTPTransport

    transport = RequestsHTTPTransport(url="http://localhost/")
    client = Client(transport=transport, fetch_schema_from_transport=True)

    try:
        with client:
            pass
    except Exception:
        # we don't care what exception is thrown, we just want to check if the
        # transport is closed afterwards
        pass

    assert isinstance(client.transport, RequestsHTTPTransport)
    assert client.transport.session is None


@pytest.mark.aiohttp
@pytest.mark.asyncio
async def test_async_transport_close_on_schema_retrieval_failure():
    """
    Ensure that the transport session is closed if an error occurs when
    entering the context manager (e.g., because schema retrieval fails)
    """

    from gql.transport.aiohttp import AIOHTTPTransport

    transport = AIOHTTPTransport(url="http://localhost/")
    client = Client(transport=transport, fetch_schema_from_transport=True)

    try:
        async with client:
            pass
    except Exception:
        # we don't care what exception is thrown, we just want to check if the
        # transport is closed afterwards
        pass

    assert isinstance(client.transport, AIOHTTPTransport)
    assert client.transport.session is None

    import asyncio

    await asyncio.sleep(1)



import json
from typing import Mapping

import pytest
from graphql import print_ast

from gql import gql
from gql.transport.exceptions import TransportConnectionFailed

# Marking all tests in this file with the websockets marker
pytestmark = pytest.mark.websockets

query1_str = """
    query getContinents {
      continents {
        code
        name
      }
    }
"""

query1_server_answer = (
    '{{"type":"data","id":"{query_id}","payload":{{"data":{{"continents":['
    '{{"code":"AF","name":"Africa"}},{{"code":"AN","name":"Antarctica"}},'
    '{{"code":"AS","name":"Asia"}},{{"code":"EU","name":"Europe"}},'
    '{{"code":"NA","name":"North America"}},{{"code":"OC","name":"Oceania"}},'
    '{{"code":"SA","name":"South America"}}]}}}}}}'
)

server1_answers = [
    query1_server_answer,
]


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server1_answers], indirect=True)
async def test_websockets_adapter_simple_query(server):
    from gql.transport.common.adapters.websockets import WebSocketsAdapter

    url = f"ws://{server.hostname}:{server.port}/graphql"

    query = print_ast(gql(query1_str).document)
    print("query=", query)

    adapter = WebSocketsAdapter(url)

    await adapter.connect()

    init_message = json.dumps({"type": "connection_init", "payload": {}})

    await adapter.send(init_message)

    result = await adapter.receive()
    print(f"result={result}")

    payload = json.dumps({"query": query})
    query_message = json.dumps({"id": 1, "type": "start", "payload": payload})

    await adapter.send(query_message)

    result = await adapter.receive()
    print(f"result={result}")

    await adapter.close()


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [server1_answers], indirect=True)
async def test_websockets_adapter_edge_cases(server):
    from gql.transport.common.adapters.websockets import WebSocketsAdapter

    url = f"ws://{server.hostname}:{server.port}/graphql"

    query = print_ast(gql(query1_str).document)
    print("query=", query)

    adapter = WebSocketsAdapter(url, headers={"a": "r1"}, ssl=False, connect_args={})

    await adapter.connect()

    assert isinstance(adapter.headers, Mapping)
    assert adapter.headers["a"] == "r1"
    assert adapter.ssl is False
    assert adapter.connect_args == {}
    assert adapter.response_headers["dummy"] == "test1234"

    # Connect twice causes AssertionError
    with pytest.raises(AssertionError):
        await adapter.connect()

    await adapter.close()

    # Second close call is ignored
    await adapter.close()

    with pytest.raises(TransportConnectionFailed):
        await adapter.send("Blah")

    with pytest.raises(TransportConnectionFailed):
        await adapter.receive()



import logging
from ssl import SSLContext
from typing import Any, Dict, Optional, Union

import websockets
from websockets import ClientConnection
from websockets.datastructures import Headers, HeadersLike

from ...exceptions import TransportConnectionFailed, TransportProtocolError
from .connection import AdapterConnection

log = logging.getLogger("gql.transport.common.adapters.websockets")


class WebSocketsAdapter(AdapterConnection):
    """AdapterConnection implementation using the websockets library."""

    def __init__(
        self,
        url: str,
        *,
        headers: Optional[HeadersLike] = None,
        ssl: Union[SSLContext, bool] = False,
        connect_args: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Initialize the transport with the given parameters.

        :param url: The GraphQL server URL. Example: 'wss://server.com:PORT/graphql'.
        :param headers: Dict of HTTP Headers.
        :param ssl: ssl_context of the connection. Use ssl=False to disable encryption
        :param connect_args: Other parameters forwarded to
            `websockets.connect <https://websockets.readthedocs.io/en/stable/reference/\
            client.html#opening-a-connection>`_
        """
        super().__init__(
            url=url,
            connect_args=connect_args,
        )

        self._headers: Optional[HeadersLike] = headers
        self.ssl = ssl

        self.websocket: Optional[ClientConnection] = None
        self._response_headers: Optional[Headers] = None

    async def connect(self) -> None:
        """Connect to the WebSocket server."""

        assert self.websocket is None

        ssl: Optional[Union[SSLContext, bool]]
        if self.ssl:
            ssl = self.ssl
        else:
            ssl = True if self.url.startswith("wss") else None

        # Set default arguments used in the websockets.connect call
        connect_args: Dict[str, Any] = {
            "ssl": ssl,
            "additional_headers": self.headers,
        }

        if self.subprotocols:
            connect_args["subprotocols"] = self.subprotocols

        # Adding custom parameters passed from init
        connect_args.update(self.connect_args)

        # Connection to the specified url
        try:
            self.websocket = await websockets.connect(self.url, **connect_args)
        except Exception as e:
            raise TransportConnectionFailed("Connect failed") from e

        assert self.websocket.response is not None

        self._response_headers = self.websocket.response.headers

    async def send(self, message: str) -> None:
        """Send message to the WebSocket server.

        Args:
            message: String message to send

        Raises:
            TransportConnectionFailed: If connection closed
        """
        if self.websocket is None:
            raise TransportConnectionFailed("WebSocket connection is already closed")

        try:
            await self.websocket.send(message)
        except Exception as e:
            raise TransportConnectionFailed(
                f"Error trying to send data: {type(e).__name__}"
            ) from e

    async def receive(self) -> str:
        """Receive message from the WebSocket server.

        Returns:
            String message received

        Raises:
            TransportConnectionFailed: If connection closed
            TransportProtocolError: If protocol error or binary data received
        """
        # It is possible that the websocket has been already closed in another task
        if self.websocket is None:
            raise TransportConnectionFailed("Connection is already closed")

        # Wait for the next websocket frame. Can raise ConnectionClosed
        try:
            data = await self.websocket.recv()
        except Exception as e:
            raise TransportConnectionFailed(
                f"Error trying to receive data: {type(e).__name__}"
            ) from e

        # websocket.recv() can return either str or bytes
        # In our case, we should receive only str here
        if not isinstance(data, str):
            raise TransportProtocolError("Binary data received in the websocket")

        answer: str = data

        return answer

    async def close(self) -> None:
        """Close the WebSocket connection."""
        if self.websocket:
            websocket = self.websocket
            self.websocket = None
            await websocket.close()

    @property
    def headers(self) -> Optional[HeadersLike]:
        """Get the response headers from the WebSocket connection.

        Returns:
            Dictionary of response headers
        """
        if self._headers:
            return self._headers
        return {}

    @property
    def response_headers(self) -> Dict[str, str]:
        """Get the response headers from the WebSocket connection.

        Returns:
            Dictionary of response headers
        """
        if self._response_headers:
            return dict(self._response_headers.raw_items())
        return {}



from typing import Any, Dict, Optional

import pytest
from graphql import (
    GraphQLArgument,
    GraphQLError,
    GraphQLField,
    GraphQLFloat,
    GraphQLInt,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLSchema,
)
from graphql.language import ValueNode
from graphql.utilities import value_from_ast_untyped

from gql import Client, gql
from gql.dsl import DSLSchema

from ..conftest import strip_braces_spaces

# Marking all tests in this file with the aiohttp marker
pytestmark = pytest.mark.aiohttp


def serialize_json(value: Any) -> Dict[str, Any]:
    return value


def parse_json_value(value: Any) -> Any:
    return value


def parse_json_literal(
    value_node: ValueNode, variables: Optional[Dict[str, Any]] = None
) -> Any:
    return value_from_ast_untyped(value_node, variables)


JsonScalar = GraphQLScalarType(
    name="JSON",
    serialize=serialize_json,
    parse_value=parse_json_value,
    parse_literal=parse_json_literal,
)

root_value = {
    "players": [
        {
            "name": "John",
            "level": 3,
            "is_connected": True,
            "score": 123.45,
            "friends": ["Alex", "Alicia"],
        },
        {
            "name": "Alex",
            "level": 4,
            "is_connected": False,
            "score": 1337.69,
            "friends": None,
        },
    ]
}


def resolve_players(root, _info):
    return root["players"]


queryType = GraphQLObjectType(
    name="Query",
    fields={"players": GraphQLField(JsonScalar, resolve=resolve_players)},
)


def resolve_add_player(root, _info, player):
    print(f"player = {player!r}")
    root["players"].append(player)
    return {"players": root["players"]}


mutationType = GraphQLObjectType(
    name="Mutation",
    fields={
        "addPlayer": GraphQLField(
            JsonScalar,
            args={"player": GraphQLArgument(GraphQLNonNull(JsonScalar))},
            resolve=resolve_add_player,
        )
    },
)

schema = GraphQLSchema(query=queryType, mutation=mutationType)


def test_json_value_output():

    client = Client(schema=schema, parse_results=True)

    query = gql("query {players}")

    result = client.execute(query, root_value=root_value)

    print(result)

    assert result["players"] == serialize_json(root_value["players"])


def test_json_value_input_in_ast():

    client = Client(schema=schema)

    query = gql(
        """
    mutation adding_player {
        addPlayer(player:  {
          name: "Tom",
          level: 1,
          is_connected: True,
          score: 0,
          friends: [
              "John"
          ]
        })
}"""
    )

    result = client.execute(query, root_value=root_value)

    print(result)

    players = result["addPlayer"]["players"]

    assert players == serialize_json(root_value["players"])
    assert players[-1]["name"] == "Tom"


def test_json_value_input_in_ast_with_variables():

    print(f"{schema.type_map!r}")
    client = Client(schema=schema)

    # Note: we need to manually add the built-in types which
    # are not present in the schema
    schema.type_map["Int"] = GraphQLInt
    schema.type_map["Float"] = GraphQLFloat

    query = gql(
        """
    mutation adding_player(
        $name: String!,
        $level: Int!,
        $is_connected: Boolean,
        $score: Float!,
        $friends: [String!]!) {

        addPlayer(player:  {
          name: $name,
          level: $level,
          is_connected: $is_connected,
          score: $score,
          friends: $friends,
        })
}"""
    )

    query.variable_values = {
        "name": "Barbara",
        "level": 1,
        "is_connected": False,
        "score": 69,
        "friends": ["Alex", "John"],
    }

    result = client.execute(query, root_value=root_value)

    print(result)

    players = result["addPlayer"]["players"]

    assert players == serialize_json(root_value["players"])
    assert players[-1]["name"] == "Barbara"


def test_json_value_input_in_dsl_argument():

    ds = DSLSchema(schema)

    new_player = {
        "name": "Tim",
        "level": 0,
        "is_connected": False,
        "score": 5,
        "friends": ["Lea"],
    }

    query = ds.Mutation.addPlayer(player=new_player)

    print(str(query))

    assert (
        strip_braces_spaces(str(query))
        == """addPlayer(
  player: {name: "Tim", level: 0, is_connected: false, score: 5, friends: ["Lea"]}
)"""
    )


def test_none_json_value_input_in_dsl_argument():

    ds = DSLSchema(schema)

    with pytest.raises(GraphQLError) as exc_info:
        ds.Mutation.addPlayer(player=None)

    assert "Received Null value for a Non-Null type JSON." in str(exc_info.value)


def test_json_value_input_with_none_list_in_dsl_argument():

    ds = DSLSchema(schema)

    new_player = {
        "name": "Bob",
        "level": 9001,
        "is_connected": True,
        "score": 666.66,
        "friends": None,
    }

    query = ds.Mutation.addPlayer(player=new_player)

    print(str(query))

    assert (
        strip_braces_spaces(str(query))
        == """addPlayer(
  player: {name: "Bob", level: 9001, is_connected: true, score: 666.66, friends: null}
)"""
    )



from enum import Enum
from typing import Any, Dict, Mapping, Type, Union, cast

from graphql import GraphQLEnumType, GraphQLSchema


def update_schema_enum(
    schema: GraphQLSchema,
    name: str,
    values: Union[Dict[str, Any], Type[Enum]],
    use_enum_values: bool = False,
) -> None:
    """Update in the schema the GraphQLEnumType corresponding to the given name.

    Example::

        from enum import Enum

        class Color(Enum):
            RED = 0
            GREEN = 1
            BLUE = 2

        update_schema_enum(schema, 'Color', Color)

    :param schema: a GraphQL Schema already containing the GraphQLEnumType type.
    :param name: the name of the enum in the GraphQL schema
    :param values: Either a Python Enum or a dict of values. The keys of the provided
        values should correspond to the keys of the existing enum in the schema.
    :param use_enum_values: By default, we configure the GraphQLEnumType to serialize
        to enum instances (ie: .parse_value() returns Color.RED).
        If use_enum_values is set to True, then .parse_value() returns 0.
        use_enum_values=True is the defaut behaviour when passing an Enum
        to a GraphQLEnumType.
    """

    # Convert Enum values to Dict
    if isinstance(values, type):
        if issubclass(values, Enum):
            values = cast(Type[Enum], values)
            if use_enum_values:
                values = {enum.name: enum.value for enum in values}
            else:
                values = {enum.name: enum for enum in values}

    if not isinstance(values, Mapping):
        raise TypeError(f"Invalid type for enum values: {type(values)}")

    # Find enum type in schema
    schema_enum = schema.get_type(name)

    if schema_enum is None:
        raise KeyError(f"Enum {name} not found in schema!")

    if not isinstance(schema_enum, GraphQLEnumType):
        raise TypeError(
            f'The type "{name}" is not a GraphQLEnumType, it is a {type(schema_enum)}'
        )

    # Replace all enum values
    for enum_name, enum_value in schema_enum.values.items():
        try:
            enum_value.value = values[enum_name]
        except KeyError:
            raise KeyError(f'Enum key "{enum_name}" not found in provided values!')

    # Delete the _value_lookup cached property
    if "_value_lookup" in schema_enum.__dict__:
        del schema_enum.__dict__["_value_lookup"]



import asyncio

import pytest

from gql import Client, gql
from gql.transport.exceptions import (
    TransportProtocolError,
    TransportQueryError,
    TransportServerError,
)

from .conftest import MS

# Marking all tests in this file with the websockets marker
pytestmark = pytest.mark.websockets


def ensure_list(s):
    return (
        s
        if s is None or isinstance(s, list)
        else list(s) if isinstance(s, tuple) else [s]
    )


query1_str = """
    query getContinents {
      continents {
        code
        name
      }
    }
"""

default_query_server_answer = (
    '{"event":"phx_reply",'
    '"payload":'
    '{"response":'
    '{"data":{"continents":['
    '{"code":"AF","name":"Africa"},{"code":"AN","name":"Antarctica"},'
    '{"code":"AS","name":"Asia"},{"code":"EU","name":"Europe"},'
    '{"code":"NA","name":"North America"},{"code":"OC","name":"Oceania"},'
    '{"code":"SA","name":"South America"}]}},'
    '"status":"ok"},'
    '"ref":2,'
    '"topic":"test_topic"}'
)


# other protocol exceptions

reply_ref_null_answer = (
    '{"event":"phx_reply","payload":{}',
    '"ref":null,' '"topic":"test_topic"}',
)

reply_ref_zero_answer = (
    '{"event":"phx_reply","payload":{}',
    '"ref":0,' '"topic":"test_topic"}',
)


# "status":"error" responses

generic_error_server_answer = (
    '{"event":"phx_reply",'
    '"payload":'
    '{"status":"error"},'
    '"ref":2,'
    '"topic":"test_topic"}'
)

error_with_reason_server_answer = (
    '{"event":"phx_reply",'
    '"payload":'
    '{"response":{"reason":"internal error"},'
    '"status":"error"},'
    '"ref":2,'
    '"topic":"test_topic"}'
)

multiple_errors_server_answer = (
    '{"event":"phx_reply",'
    '"payload":'
    '{"response":{"errors": ["error 1", "error 2"]},'
    '"status":"error"},'
    '"ref":2,'
    '"topic":"test_topic"}'
)

timeout_server_answer = (
    '{"event":"phx_reply",'
    '"payload":'
    '{"status":"timeout"},'
    '"ref":2,'
    '"topic":"test_topic"}'
)

invalid_payload_data_answer = (
    '{"event":"phx_reply",' '"payload":"INVALID",' '"ref":2,' '"topic":"test_topic"}'
)

# "status":"ok" exceptions

invalid_response_server_answer = (
    '{"event":"phx_reply",'
    '"payload":{"response":"INVALID",'
    '"status":"ok"}'
    '"ref":2,'
    '"topic":"test_topic"}'
)

invalid_response_keys_server_answer = (
    '{"event":"phx_reply",'
    '"payload":{"response":'
    '{"data":{"continents":null},"invalid":null}",'
    '"status":"ok"}'
    '"ref":2,'
    '"topic":"test_topic"}'
)

invalid_event_server_answer = '{"event":"unknown"}'


def query_server(server_answers=default_query_server_answer):
    from .conftest import PhoenixChannelServerHelper

    async def phoenix_server(ws):
        await PhoenixChannelServerHelper.send_connection_ack(ws)
        await ws.recv()
        for server_answer in ensure_list(server_answers):
            await ws.send(server_answer)
        await PhoenixChannelServerHelper.send_close(ws)
        await ws.wait_closed()

    return phoenix_server


async def no_connection_ack_phoenix_server(ws):
    from .conftest import PhoenixChannelServerHelper

    await ws.recv()
    await PhoenixChannelServerHelper.send_close(ws)
    await ws.wait_closed()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "server",
    [
        query_server(reply_ref_null_answer),
        query_server(reply_ref_zero_answer),
        query_server(invalid_payload_data_answer),
        query_server(invalid_response_server_answer),
        query_server(invalid_response_keys_server_answer),
        no_connection_ack_phoenix_server,
        query_server(invalid_event_server_answer),
    ],
    indirect=True,
)
@pytest.mark.parametrize("query_str", [query1_str])
async def test_phoenix_channel_query_protocol_error(server, query_str):

    from gql.transport.phoenix_channel_websockets import (
        PhoenixChannelWebsocketsTransport,
    )

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"
    transport = PhoenixChannelWebsocketsTransport(channel_name="test_channel", url=url)

    query = gql(query_str)
    with pytest.raises(TransportProtocolError):
        async with Client(transport=transport) as session:
            await session.execute(query)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "server",
    [
        query_server(generic_error_server_answer),
        query_server(error_with_reason_server_answer),
        query_server(multiple_errors_server_answer),
        query_server(timeout_server_answer),
    ],
    indirect=True,
)
@pytest.mark.parametrize("query_str", [query1_str])
async def test_phoenix_channel_query_error(server, query_str):

    from gql.transport.phoenix_channel_websockets import (
        PhoenixChannelWebsocketsTransport,
    )

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"
    transport = PhoenixChannelWebsocketsTransport(channel_name="test_channel", url=url)

    query = gql(query_str)
    with pytest.raises(TransportQueryError):
        async with Client(transport=transport) as session:
            await session.execute(query)


query2_str = """
    subscription getContinents {
      continents {
        code
        name
      }
    }
"""

default_subscription_server_answer = (
    '{"event":"phx_reply",'
    '"payload":'
    '{"response":{"subscriptionId":"test_subscription"},'
    '"status":"ok"},'
    '"ref":2,'
    '"topic":"test_topic"}'
)

ref_is_not_an_integer_server_answer = (
    '{"event":"phx_reply",'
    '"payload":'
    '{"response":{"subscriptionId":"test_subscription"},'
    '"status":"ok"},'
    '"ref":"not_an_integer",'
    '"topic":"test_topic"}'
)

missing_ref_server_answer = (
    '{"event":"phx_reply",'
    '"payload":'
    '{"response":{"subscriptionId":"test_subscription"},'
    '"status":"ok"},'
    '"topic":"test_topic"}'
)

missing_subscription_id_server_answer = (
    '{"event":"phx_reply",'
    '"payload":'
    '{"response":{},"status":"ok"},'
    '"ref":2,'
    '"topic":"test_topic"}'
)

null_subscription_id_server_answer = (
    '{"event":"phx_reply",'
    '"payload":'
    '{"response":{"subscriptionId":null},"status":"ok"},'
    '"ref":2,'
    '"topic":"test_topic"}'
)

default_subscription_data_answer = (
    '{"event":"subscription:data","payload":'
    '{"subscriptionId":"test_subscription","result":'
    '{"data":{"continents":['
    '{"code":"AF","name":"Africa"},{"code":"AN","name":"Antarctica"},'
    '{"code":"AS","name":"Asia"},{"code":"EU","name":"Europe"},'
    '{"code":"NA","name":"North America"},{"code":"OC","name":"Oceania"},'
    '{"code":"SA","name":"South America"}]}}},'
    '"ref":null,'
    '"topic":"test_subscription"}'
)

default_subscription_unsubscribe_answer = (
    '{"event":"phx_reply",'
    '"payload":{"response":{"subscriptionId":"test_subscription"},'
    '"status":"ok"},'
    '"ref":3,'
    '"topic":"test_topic"}'
)

missing_subscription_id_data_answer = (
    '{"event":"subscription:data","payload":'
    '{"result":'
    '{"data":{"continents":['
    '{"code":"AF","name":"Africa"},{"code":"AN","name":"Antarctica"},'
    '{"code":"AS","name":"Asia"},{"code":"EU","name":"Europe"},'
    '{"code":"NA","name":"North America"},{"code":"OC","name":"Oceania"},'
    '{"code":"SA","name":"South America"}]}}},'
    '"ref":null,'
    '"topic":"test_subscription"}'
)

null_subscription_id_data_answer = (
    '{"event":"subscription:data","payload":'
    '{"subscriptionId":null,"result":'
    '{"data":{"continents":['
    '{"code":"AF","name":"Africa"},{"code":"AN","name":"Antarctica"},'
    '{"code":"AS","name":"Asia"},{"code":"EU","name":"Europe"},'
    '{"code":"NA","name":"North America"},{"code":"OC","name":"Oceania"},'
    '{"code":"SA","name":"South America"}]}}},'
    '"ref":null,'
    '"topic":"test_subscription"}'
)

invalid_subscription_id_data_answer = (
    '{"event":"subscription:data","payload":'
    '{"subscriptionId":"INVALID","result":'
    '{"data":{"continents":['
    '{"code":"AF","name":"Africa"},{"code":"AN","name":"Antarctica"},'
    '{"code":"AS","name":"Asia"},{"code":"EU","name":"Europe"},'
    '{"code":"NA","name":"North America"},{"code":"OC","name":"Oceania"},'
    '{"code":"SA","name":"South America"}]}}},'
    '"ref":null,'
    '"topic":"test_subscription"}'
)

invalid_payload_data_answer = (
    '{"event":"subscription:data",'
    '"payload":"INVALID",'
    '"ref":null,'
    '"topic":"test_subscription"}'
)

invalid_result_data_answer = (
    '{"event":"subscription:data","payload":'
    '{"subscriptionId":"test_subscription","result":"INVALID"},'
    '"ref":null,'
    '"topic":"test_subscription"}'
)

invalid_result_keys_data_answer = (
    '{"event":"subscription:data",'
    '"payload":{"subscriptionId":"test_subscription",'
    '"result":{"data":{"continents":null},"invalid":null}},'
    '"ref":null,'
    '"topic":"test_subscription"}'
)

invalid_subscription_ref_answer = (
    '{"event":"phx_reply",'
    '"payload":{"response":{"subscriptionId":"test_subscription"},'
    '"status":"ok"},'
    '"ref":99,'
    '"topic":"test_topic"}'
)

mismatched_unsubscribe_answer = (
    '{"event":"phx_reply",'
    '"payload":{"response":{"subscriptionId":"no_such_subscription"},'
    '"status":"ok"},'
    '"ref":3,'
    '"topic":"test_topic"}'
)


def subscription_server(
    server_answers=default_subscription_server_answer,
    data_answers=default_subscription_data_answer,
    unsubscribe_answers=default_subscription_unsubscribe_answer,
):
    import json

    from .conftest import PhoenixChannelServerHelper

    async def phoenix_server(ws):
        await PhoenixChannelServerHelper.send_connection_ack(ws)
        await ws.recv()
        if server_answers is not None:
            for server_answer in ensure_list(server_answers):
                await ws.send(server_answer)
        if data_answers is not None:
            for data_answer in ensure_list(data_answers):
                await ws.send(data_answer)
        if unsubscribe_answers is not None:
            result = await ws.recv()
            json_result = json.loads(result)
            assert json_result["event"] == "unsubscribe"
            for unsubscribe_answer in ensure_list(unsubscribe_answers):
                await ws.send(unsubscribe_answer)
        else:
            await PhoenixChannelServerHelper.send_close(ws)
        await ws.wait_closed()

    return phoenix_server


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "server",
    [
        subscription_server(invalid_subscription_ref_answer),
        subscription_server(missing_subscription_id_server_answer),
        subscription_server(null_subscription_id_server_answer),
        subscription_server(
            [default_subscription_server_answer, default_subscription_server_answer]
        ),
        subscription_server(data_answers=missing_subscription_id_data_answer),
        subscription_server(data_answers=null_subscription_id_data_answer),
        subscription_server(data_answers=invalid_subscription_id_data_answer),
        subscription_server(data_answers=ref_is_not_an_integer_server_answer),
        subscription_server(data_answers=missing_ref_server_answer),
        subscription_server(data_answers=invalid_payload_data_answer),
        subscription_server(data_answers=invalid_result_data_answer),
        subscription_server(data_answers=invalid_result_keys_data_answer),
    ],
    indirect=True,
)
@pytest.mark.parametrize("query_str", [query2_str])
async def test_phoenix_channel_subscription_protocol_error(server, query_str):

    from gql.transport.phoenix_channel_websockets import (
        PhoenixChannelWebsocketsTransport,
    )

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"
    transport = PhoenixChannelWebsocketsTransport(channel_name="test_channel", url=url)

    query = gql(query_str)
    with pytest.raises(TransportProtocolError):
        async with Client(transport=transport) as session:
            async for _result in session.subscribe(query):
                await asyncio.sleep(10 * MS)
                break


server_error_server_answer = '{"event":"phx_error", "ref":2, "topic":"test_topic"}'


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "server",
    [query_server(server_error_server_answer)],
    indirect=True,
)
@pytest.mark.parametrize("query_str", [query1_str])
async def test_phoenix_channel_server_error(server, query_str):

    from gql.transport.phoenix_channel_websockets import (
        PhoenixChannelWebsocketsTransport,
    )

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"
    transport = PhoenixChannelWebsocketsTransport(channel_name="test_channel", url=url)

    query = gql(query_str)
    with pytest.raises(TransportServerError):
        async with Client(transport=transport) as session:
            await session.execute(query)


# These cannot be caught by the client
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "server",
    [
        subscription_server(unsubscribe_answers=invalid_subscription_ref_answer),
        subscription_server(unsubscribe_answers=mismatched_unsubscribe_answer),
    ],
    indirect=True,
)
@pytest.mark.parametrize("query_str", [query2_str])
async def test_phoenix_channel_unsubscribe_error(server, query_str):

    from gql.transport.phoenix_channel_websockets import (
        PhoenixChannelWebsocketsTransport,
    )

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"

    # Reduce close_timeout. These tests will wait for an unsubscribe
    # reply that will never come...
    transport = PhoenixChannelWebsocketsTransport(
        channel_name="test_channel", url=url, close_timeout=1
    )

    query = gql(query_str)
    async with Client(transport=transport) as session:
        async for _result in session.subscribe(query):
            break


# We can force the error if somehow the generator is still running while
# we receive a mismatched unsubscribe answer
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "server",
    [subscription_server(unsubscribe_answers=mismatched_unsubscribe_answer)],
    indirect=True,
)
@pytest.mark.parametrize("query_str", [query2_str])
async def test_phoenix_channel_unsubscribe_error_forcing(server, query_str):

    from gql.transport.phoenix_channel_websockets import (
        PhoenixChannelWebsocketsTransport,
    )

    path = "/graphql"
    url = f"ws://{server.hostname}:{server.port}{path}"

    transport = PhoenixChannelWebsocketsTransport(
        channel_name="test_channel", url=url, close_timeout=1
    )

    query = gql(query_str)
    with pytest.raises(TransportProtocolError):
        async with Client(transport=transport) as session:
            async for _result in session.subscribe(query):
                await session.transport._send_stop_message(2)
                await asyncio.sleep(10 * MS)



import asyncio
import json
import logging
from typing import Any, Dict, Optional, Tuple, Union

from graphql import ExecutionResult, print_ast

from ..graphql_request import GraphQLRequest
from .common.adapters.websockets import WebSocketsAdapter
from .common.base import SubscriptionTransportBase
from .exceptions import (
    TransportConnectionFailed,
    TransportProtocolError,
    TransportQueryError,
    TransportServerError,
)

log = logging.getLogger(__name__)


class Subscription:
    """Records listener_id and unsubscribe query_id for a subscription."""

    def __init__(self, query_id: int) -> None:
        self.listener_id: int = query_id
        self.unsubscribe_id: Optional[int] = None


class PhoenixChannelWebsocketsTransport(SubscriptionTransportBase):
    """The PhoenixChannelWebsocketsTransport is an async transport
    which allows you to execute queries and subscriptions against an `Absinthe`_
    backend using the `Phoenix`_ framework `channels`_.

    .. _Absinthe: http://absinthe-graphql.org
    .. _Phoenix: https://www.phoenixframework.org
    .. _channels: https://hexdocs.pm/phoenix/Phoenix.Channel.html#content
    """

    def __init__(
        self,
        url: str,
        *,
        channel_name: str = "__absinthe__:control",
        heartbeat_interval: float = 30,
        ack_timeout: Optional[Union[int, float]] = 10,
        **kwargs: Any,
    ) -> None:
        """Initialize the transport with the given parameters.

        :param url: The server URL.'.
        :param channel_name: Channel on the server this transport will join.
            The default for Absinthe servers is "__absinthe__:control"
        :param heartbeat_interval: Interval in second between each heartbeat messages
            sent by the client
        :param ack_timeout: Timeout in seconds to wait for the reply message
            from the server.
        """
        self.channel_name: str = channel_name
        self.heartbeat_interval: float = heartbeat_interval
        self.heartbeat_task: Optional[asyncio.Future] = None
        self.subscriptions: Dict[str, Subscription] = {}
        self.ack_timeout: Optional[Union[int, float]] = ack_timeout

        # Instanciate a WebSocketAdapter to indicate the use
        # of the websockets dependency for this transport
        ws_adapter_args = {}
        for ws_arg in ["headers", "ssl", "connect_args"]:
            try:
                ws_adapter_args[ws_arg] = kwargs.pop(ws_arg)
            except KeyError:
                pass

        self.adapter: WebSocketsAdapter = WebSocketsAdapter(
            url=url,
            **ws_adapter_args,
        )

        # Initialize the generic SubscriptionTransportBase parent class
        super().__init__(
            adapter=self.adapter,
            **kwargs,
        )

    async def _initialize(self) -> None:
        """Join the specified channel and wait for the connection ACK.

        If the answer is not a connection_ack message, we will return an Exception.
        """

        query_id = self.next_query_id
        self.next_query_id += 1

        init_message = json.dumps(
            {
                "topic": self.channel_name,
                "event": "phx_join",
                "payload": {},
                "ref": query_id,
            }
        )

        await self._send(init_message)

        # Wait for the connection_ack message or raise a TimeoutError
        init_answer = await asyncio.wait_for(self._receive(), self.ack_timeout)

        answer_type, answer_id, execution_result = self._parse_answer(init_answer)

        if answer_type != "reply":
            raise TransportProtocolError(
                "Websocket server did not return a connection ack"
            )

        async def heartbeat_coro():
            while True:
                await asyncio.sleep(self.heartbeat_interval)
                try:
                    query_id = self.next_query_id
                    self.next_query_id += 1

                    await self._send(
                        json.dumps(
                            {
                                "topic": "phoenix",
                                "event": "heartbeat",
                                "payload": {},
                                "ref": query_id,
                            }
                        )
                    )
                except TransportConnectionFailed:  # pragma: no cover
                    return

        self.heartbeat_task = asyncio.ensure_future(heartbeat_coro())

    async def _send_stop_message(self, query_id: int) -> None:
        """Send an 'unsubscribe' message to the Phoenix Channel referencing
        the listener's query_id, saving the query_id of the message.

        The server should afterwards return a 'phx_reply' message with
        the same query_id and subscription_id of the 'unsubscribe' request.
        """
        subscription_id = self._find_existing_subscription(query_id)

        unsubscribe_query_id = self.next_query_id
        self.next_query_id += 1

        # Save the ref so it can be matched in the reply
        self.subscriptions[subscription_id].unsubscribe_id = unsubscribe_query_id
        unsubscribe_message = json.dumps(
            {
                "topic": self.channel_name,
                "event": "unsubscribe",
                "payload": {"subscriptionId": subscription_id},
                "ref": unsubscribe_query_id,
            }
        )

        await self._send(unsubscribe_message)

    async def _stop_listener(self, query_id: int) -> None:
        await self._send_stop_message(query_id)

    async def _send_connection_terminate_message(self) -> None:
        """Send a phx_leave message to disconnect from the provided channel."""

        query_id = self.next_query_id
        self.next_query_id += 1

        connection_terminate_message = json.dumps(
            {
                "topic": self.channel_name,
                "event": "phx_leave",
                "payload": {},
                "ref": query_id,
            }
        )

        await self._send(connection_terminate_message)

    async def _connection_terminate(self):
        await self._send_connection_terminate_message()

    async def _send_query(
        self,
        request: GraphQLRequest,
    ) -> int:
        """Send a query to the provided websocket connection.

        We use an incremented id to reference the query.

        Returns the used id for this query.
        """

        query_id = self.next_query_id
        self.next_query_id += 1

        query_str = json.dumps(
            {
                "topic": self.channel_name,
                "event": "doc",
                "payload": {
                    "query": print_ast(request.document),
                    "variables": request.variable_values or {},
                },
                "ref": query_id,
            }
        )

        await self._send(query_str)

        return query_id

    def _parse_answer(
        self, answer: str
    ) -> Tuple[str, Optional[int], Optional[ExecutionResult]]:
        """Parse the answer received from the server

        Returns a list consisting of:
            - the answer_type (between:
              'data', 'reply', 'complete', 'close')
            - the answer id (Integer) if received or None
            - an execution Result if the answer_type is 'data' or None
        """

        event: str = ""
        answer_id: Optional[int] = None
        answer_type: str = ""
        execution_result: Optional[ExecutionResult] = None
        subscription_id: Optional[str] = None

        def _get_value(d: Any, key: str, label: str) -> Any:
            if not isinstance(d, dict):
                raise ValueError(f"{label} is not a dict")

            return d.get(key)

        def _required_value(d: Any, key: str, label: str) -> Any:
            value = _get_value(d, key, label)
            if value is None:
                raise ValueError(f"null {key} in {label}")

            return value

        def _required_subscription_id(
            d: Any, label: str, must_exist: bool = False, must_not_exist: bool = False
        ) -> str:
            subscription_id = str(_required_value(d, "subscriptionId", label))
            if must_exist and (subscription_id not in self.subscriptions):
                raise ValueError("unregistered subscriptionId")
            if must_not_exist and (subscription_id in self.subscriptions):
                raise ValueError("previously registered subscriptionId")

            return subscription_id

        def _validate_data_response(d: Any, label: str) -> dict:
            """Make sure query, mutation or subscription answer conforms.
            The GraphQL spec says only three keys are permitted.
            """
            if not isinstance(d, dict):
                raise ValueError(f"{label} is not a dict")

            keys = set(d.keys())
            invalid = keys - {"data", "errors", "extensions"}
            if len(invalid) > 0:
                raise ValueError(
                    f"{label} contains invalid items: " + ", ".join(invalid)
                )
            return d

        try:
            json_answer = json.loads(answer)

            event = str(_required_value(json_answer, "event", "answer"))

            if event == "subscription:data":
                payload = _required_value(json_answer, "payload", "answer")

                subscription_id = _required_subscription_id(
                    payload, "payload", must_exist=True
                )

                result = _validate_data_response(payload.get("result"), "result")

                answer_type = "data"

                subscription = self.subscriptions[subscription_id]
                answer_id = subscription.listener_id

                execution_result = ExecutionResult(
                    data=result.get("data"),
                    errors=result.get("errors"),
                    extensions=result.get("extensions"),
                )

            elif event == "phx_reply":

                # Will generate a ValueError if 'ref' is not there
                # or if it is not an integer
                answer_id = int(_required_value(json_answer, "ref", "answer"))

                payload = _required_value(json_answer, "payload", "answer")

                status = _get_value(payload, "status", "payload")

                if status == "ok":
                    answer_type = "reply"

                    if answer_id in self.listeners:
                        response = _required_value(payload, "response", "payload")

                        if isinstance(response, dict) and "subscriptionId" in response:

                            # Subscription answer
                            subscription_id = _required_subscription_id(
                                response, "response", must_not_exist=True
                            )

                            self.subscriptions[subscription_id] = Subscription(
                                answer_id
                            )

                        else:
                            # Query or mutation answer
                            # GraphQL spec says only three keys are permitted
                            response = _validate_data_response(response, "response")

                            answer_type = "data"

                            execution_result = ExecutionResult(
                                data=response.get("data"),
                                errors=response.get("errors"),
                                extensions=response.get("extensions"),
                            )
                    else:
                        (
                            registered_subscription_id,
                            listener_id,
                        ) = self._find_subscription(answer_id)
                        if registered_subscription_id is not None:
                            # Unsubscription answer
                            response = _required_value(payload, "response", "payload")
                            subscription_id = _required_subscription_id(
                                response, "response"
                            )

                            if subscription_id != registered_subscription_id:
                                raise ValueError("subscription id does not match")

                            answer_type = "complete"

                            answer_id = listener_id

                elif status == "error":
                    response = payload.get("response")

                    if isinstance(response, dict):
                        if "errors" in response:
                            raise TransportQueryError(
                                str(response.get("errors")), query_id=answer_id
                            )
                        elif "reason" in response:
                            raise TransportQueryError(
                                str(response.get("reason")), query_id=answer_id
                            )
                    raise TransportQueryError("reply error", query_id=answer_id)

                elif status == "timeout":
                    raise TransportQueryError("reply timeout", query_id=answer_id)

                # In case of missing or unrecognized status, just continue

            elif event == "phx_error":
                # Sent if the channel has crashed
                # answer_id will be the "join_ref" for the channel
                # answer_id = int(json_answer.get("ref"))
                raise TransportServerError("Server error")
            elif event == "phx_close":
                answer_type = "close"
            else:
                raise ValueError("unrecognized event")

        except ValueError as e:
            log.error(f"Error parsing answer '{answer}': {e!r}")
            raise TransportProtocolError(
                f"Server did not return a GraphQL result: {e!s}"
            ) from e

        return answer_type, answer_id, execution_result

    async def _handle_answer(
        self,
        answer_type: str,
        answer_id: Optional[int],
        execution_result: Optional[ExecutionResult],
    ) -> None:
        if answer_type == "close":
            pass
        else:
            await super()._handle_answer(answer_type, answer_id, execution_result)

    def _remove_listener(self, query_id: int) -> None:
        """If the listener was a subscription, remove that information."""
        try:
            subscription_id = self._find_existing_subscription(query_id)
            del self.subscriptions[subscription_id]
        except Exception:
            pass
        super()._remove_listener(query_id)

    def _find_subscription(self, query_id: int) -> Tuple[Optional[str], int]:
        """Perform a reverse lookup to find the subscription id matching
        a listener's query_id.
        """
        for subscription_id, subscription in self.subscriptions.items():
            if query_id == subscription.listener_id:
                return subscription_id, query_id
            if query_id == subscription.unsubscribe_id:
                return subscription_id, subscription.listener_id
        return None, query_id

    def _find_existing_subscription(self, query_id: int) -> str:
        """Perform a reverse lookup to find the subscription id matching
        a listener's query_id.
        """
        subscription_id, _listener_id = self._find_subscription(query_id)

        if subscription_id is None:
            raise TransportProtocolError(
                f"No subscription registered for listener {query_id}"
            )
        return subscription_id

    async def _close_coro(self, e: Exception, clean_close: bool = True) -> None:
        if self.heartbeat_task is not None:
            self.heartbeat_task.cancel()

        await super()._close_coro(e, clean_close)



import pytest
from graphql import (
    FloatValueNode,
    GraphQLError,
    GraphQLFloat,
    GraphQLID,
    GraphQLInputObjectType,
    GraphQLInt,
    GraphQLList,
    GraphQLNonNull,
    IntValueNode,
    ListTypeNode,
    NamedTypeNode,
    NameNode,
    NonNullTypeNode,
    NullValueNode,
    Undefined,
    build_ast_schema,
    parse,
    print_ast,
)
from graphql.utilities import get_introspection_query

from gql import Client, gql
from gql.dsl import (
    DSLFragment,
    DSLInlineFragment,
    DSLMetaField,
    DSLMutation,
    DSLQuery,
    DSLSchema,
    DSLSubscription,
    DSLVariable,
    DSLVariableDefinitions,
    ast_from_serialized_value_untyped,
    ast_from_value,
    dsl_gql,
)
from gql.utilities import get_introspection_query_ast, node_tree

from ..conftest import strip_braces_spaces
from .schema import StarWarsSchema


@pytest.fixture
def ds():
    return DSLSchema(StarWarsSchema)


@pytest.fixture
def client():
    return Client(schema=StarWarsSchema)


def test_ast_from_value_with_input_type_and_not_mapping_value():
    obj_type = StarWarsSchema.get_type("ReviewInput")
    assert isinstance(obj_type, GraphQLInputObjectType)
    assert ast_from_value(8, obj_type) is None


def test_ast_from_value_with_list_type_and_non_iterable_value():
    assert ast_from_value(5, GraphQLList(GraphQLInt)) == IntValueNode(value="5")


def test_ast_from_value_with_none():
    assert ast_from_value(None, GraphQLInt) == NullValueNode()


def test_ast_from_value_with_undefined():
    with pytest.raises(GraphQLError) as exc_info:
        ast_from_value(Undefined, GraphQLInt)

    assert "Received Undefined value for type Int." in str(exc_info.value)


def test_ast_from_value_with_graphqlid():

    assert ast_from_value("12345", GraphQLID) == IntValueNode(value="12345")


def test_ast_from_value_with_invalid_type():
    with pytest.raises(TypeError) as exc_info:
        ast_from_value(4, None)  # type: ignore

    assert "Unexpected input type: None." in str(exc_info.value)


def test_ast_from_value_with_non_null_type_and_none():
    typ = GraphQLNonNull(GraphQLInt)

    with pytest.raises(GraphQLError) as exc_info:
        ast_from_value(None, typ)

    assert "Received Null value for a Non-Null type Int." in str(exc_info.value)


def test_ast_from_value_float_precision():

    # Checking precision of float serialization
    # See https://github.com/graphql-python/graphql-core/pull/164

    assert ast_from_value(123456789.01234567, GraphQLFloat) == FloatValueNode(
        value="123456789.01234567"
    )

    assert ast_from_value(1.1, GraphQLFloat) == FloatValueNode(value="1.1")

    assert ast_from_value(123.0, GraphQLFloat) == FloatValueNode(value="123")


def test_ast_from_serialized_value_untyped_typeerror():
    with pytest.raises(TypeError) as exc_info:
        ast_from_serialized_value_untyped(GraphQLInt)

    assert "Cannot convert value to AST: Int." in str(exc_info.value)


def test_variable_to_ast_type_passing_wrapping_type():
    review_type = StarWarsSchema.get_type("ReviewInput")
    assert isinstance(review_type, GraphQLInputObjectType)

    wrapping_type = GraphQLNonNull(GraphQLList(review_type))
    variable = DSLVariable("review_input")
    ast = variable.to_ast_type(wrapping_type)
    assert ast == NonNullTypeNode(
        type=ListTypeNode(type=NamedTypeNode(name=NameNode(value="ReviewInput")))
    )


def test_use_variable_definition_multiple_times(ds):
    var = DSLVariableDefinitions()

    # `episode` variable is used in both fields
    op = DSLMutation(
        ds.Mutation.createReview.alias("badReview")
        .args(review=var.badReview, episode=var.episode)
        .select(ds.Review.stars, ds.Review.commentary),
        ds.Mutation.createReview.alias("goodReview")
        .args(review=var.goodReview, episode=var.episode)
        .select(ds.Review.stars, ds.Review.commentary),
    )
    op.variable_definitions = var
    query = dsl_gql(op)

    assert (
        print_ast(query.document)
        == """mutation \
($badReview: ReviewInput, $episode: Episode, $goodReview: ReviewInput) {
  badReview: createReview(review: $badReview, episode: $episode) {
    stars
    commentary
  }
  goodReview: createReview(review: $goodReview, episode: $episode) {
    stars
    commentary
  }
}"""
    )

    assert node_tree(query.document) == node_tree(
        gql(print_ast(query.document)).document
    )


def test_add_variable_definitions(ds):
    var = DSLVariableDefinitions()
    op = DSLMutation(
        ds.Mutation.createReview.args(review=var.review, episode=var.episode).select(
            ds.Review.stars, ds.Review.commentary
        )
    )
    op.variable_definitions = var
    query = dsl_gql(op)

    assert (
        print_ast(query.document)
        == """mutation ($review: ReviewInput, $episode: Episode) {
  createReview(review: $review, episode: $episode) {
    stars
    commentary
  }
}"""
    )

    assert node_tree(query.document) == node_tree(
        gql(print_ast(query.document)).document
    )


def test_add_variable_definitions_with_default_value_enum(ds):
    var = DSLVariableDefinitions()
    op = DSLMutation(
        ds.Mutation.createReview.args(
            review=var.review, episode=var.episode.default(4)
        ).select(ds.Review.stars, ds.Review.commentary)
    )
    op.variable_definitions = var
    query = dsl_gql(op)

    assert (
        print_ast(query.document)
        == """mutation ($review: ReviewInput, $episode: Episode = NEWHOPE) {
  createReview(review: $review, episode: $episode) {
    stars
    commentary
  }
}"""
    )


def test_add_variable_definitions_with_default_value_input_object(ds):
    var = DSLVariableDefinitions()
    op = DSLMutation(
        ds.Mutation.createReview.args(
            review=var.review.default({"stars": 5, "commentary": "Wow!"}),
            episode=var.episode,
        ).select(ds.Review.stars, ds.Review.commentary)
    )
    op.variable_definitions = var
    query = dsl_gql(op)

    assert (
        strip_braces_spaces(print_ast(query.document))
        == """
mutation ($review: ReviewInput = {stars: 5, commentary: "Wow!"}, $episode: Episode) {
  createReview(review: $review, episode: $episode) {
    stars
    commentary
  }
}""".strip()
    )

    assert node_tree(query.document) == node_tree(
        gql(print_ast(query.document)).document
    )


def test_add_variable_definitions_in_input_object(ds):
    var = DSLVariableDefinitions()
    op = DSLMutation(
        ds.Mutation.createReview.args(
            review={"stars": var.stars, "commentary": var.commentary},
            episode=var.episode,
        ).select(ds.Review.stars, ds.Review.commentary)
    )
    op.variable_definitions = var
    query = dsl_gql(op)

    assert (
        strip_braces_spaces(print_ast(query.document))
        == """mutation ($stars: Int, $commentary: String, $episode: Episode) {
  createReview(
    review: {stars: $stars, commentary: $commentary}
    episode: $episode
  ) {
    stars
    commentary
  }
}"""
    )

    assert node_tree(query.document) == node_tree(
        gql(print_ast(query.document)).document
    )


def test_invalid_field_on_type_query(ds):
    with pytest.raises(AttributeError) as exc_info:
        ds.Query.extras.select(ds.Character.name)
    assert "Field extras does not exist in type Query." in str(exc_info.value)


def test_incompatible_field(ds):
    with pytest.raises(TypeError) as exc_info:
        ds.Query.hero.select("not_a_DSL_FIELD")
    assert (
        "Fields should be instances of DSLSelectable. Received: <class 'str'>"
        in str(exc_info.value)
    )


def test_hero_name_query(ds):
    query = """
hero {
  name
}
    """.strip()
    query_dsl = ds.Query.hero.select(ds.Character.name)
    assert query == str(query_dsl)


def test_hero_name_and_friends_query(ds):
    query = """
hero {
  id
  name
  friends {
    name
  }
}
    """.strip()
    query_dsl = ds.Query.hero.select(
        ds.Character.id,
        ds.Character.name,
        ds.Character.friends.select(
            ds.Character.name,
        ),
    )
    assert query == str(query_dsl)

    # Should also work with a chain of selects
    query_dsl = (
        ds.Query.hero.select(ds.Character.id)
        .select(ds.Character.name)
        .select(
            ds.Character.friends.select(
                ds.Character.name,
            ),
        )
    )
    assert query == str(query_dsl)


def test_hero_id_and_name(ds):
    query = """
hero {
  id
  name
}
    """.strip()
    query_dsl = ds.Query.hero.select(ds.Character.id)
    query_dsl = query_dsl.select(ds.Character.name)
    assert query == str(query_dsl)


def test_nested_query(ds):
    query = """
hero {
  name
  friends {
    name
    appearsIn
    friends {
      name
    }
  }
}
    """.strip()
    query_dsl = ds.Query.hero.select(
        ds.Character.name,
        ds.Character.friends.select(
            ds.Character.name,
            ds.Character.appears_in,
            ds.Character.friends.select(ds.Character.name),
        ),
    )
    assert query == str(query_dsl)


def test_fetch_luke_query(ds):
    query = """
human(id: "1000") {
  name
}
    """.strip()
    query_dsl = ds.Query.human(id="1000").select(
        ds.Human.name,
    )

    assert query == str(query_dsl)


def test_fetch_luke_aliased(ds):
    query = """
luke: human(id: "1000") {
  name
}
    """.strip()
    query_dsl = (
        ds.Query.human.args(id=1000)
        .alias("luke")
        .select(
            ds.Character.name,
        )
    )
    assert query == str(query_dsl)

    # Should also work with select before alias
    query_dsl = (
        ds.Query.human.args(id=1000)
        .select(
            ds.Character.name,
        )
        .alias("luke")
    )
    assert query == str(query_dsl)


def test_fetch_name_aliased(ds: DSLSchema) -> None:
    query = """
human(id: "1000") {
  my_name: name
}
    """.strip()
    query_dsl = ds.Query.human.args(id=1000).select(ds.Character.name.alias("my_name"))
    print(str(query_dsl))
    assert query == str(query_dsl)


def test_fetch_name_aliased_as_kwargs(ds: DSLSchema) -> None:
    query = """
human(id: "1000") {
  my_name: name
}
    """.strip()
    query_dsl = ds.Query.human.args(id=1000).select(
        my_name=ds.Character.name,
    )
    assert query == str(query_dsl)


def test_hero_name_query_result(ds, client):
    query = dsl_gql(DSLQuery(ds.Query.hero.select(ds.Character.name)))
    result = client.execute(query)
    expected = {"hero": {"name": "R2-D2"}}
    assert result == expected
    assert node_tree(query.document) == node_tree(
        gql(print_ast(query.document)).document
    )


def test_arg_serializer_list(ds, client):
    query = dsl_gql(
        DSLQuery(
            ds.Query.characters.args(ids=[1000, 1001, 1003]).select(
                ds.Character.name,
            )
        )
    )
    result = client.execute(query)
    expected = {
        "characters": [
            {"name": "Luke Skywalker"},
            {"name": "Darth Vader"},
            {"name": "Leia Organa"},
        ]
    }
    assert result == expected
    assert node_tree(query.document) == node_tree(
        gql(print_ast(query.document)).document
    )


def test_arg_serializer_enum(ds, client):
    query = dsl_gql(DSLQuery(ds.Query.hero.args(episode=5).select(ds.Character.name)))
    result = client.execute(query)
    expected = {"hero": {"name": "Luke Skywalker"}}
    assert result == expected
    assert node_tree(query.document) == node_tree(
        gql(print_ast(query.document)).document
    )


def test_create_review_mutation_result(ds, client):

    query = dsl_gql(
        DSLMutation(
            ds.Mutation.createReview.args(
                episode=6, review={"stars": 5, "commentary": "This is a great movie!"}
            ).select(ds.Review.stars, ds.Review.commentary)
        )
    )
    result = client.execute(query)
    expected = {"createReview": {"stars": 5, "commentary": "This is a great movie!"}}
    assert result == expected
    assert node_tree(query.document) == node_tree(
        gql(print_ast(query.document)).document
    )


def test_subscription(ds):

    query = dsl_gql(
        DSLSubscription(
            ds.Subscription.reviewAdded(episode=6).select(
                ds.Review.stars, ds.Review.commentary
            )
        )
    )
    assert (
        print_ast(query.document)
        == """subscription {
  reviewAdded(episode: JEDI) {
    stars
    commentary
  }
}"""
    )

    assert node_tree(query.document) == node_tree(
        gql(print_ast(query.document)).document
    )


def test_field_does_not_exit_in_type(ds):
    with pytest.raises(
        GraphQLError,
        match="Invalid field for <DSLField Query::hero>: <DSLField Query::hero>",
    ):
        ds.Query.hero.select(ds.Query.hero)


def test_try_to_select_on_scalar_field(ds):
    with pytest.raises(
        GraphQLError,
        match="Invalid field for <DSLField Human::name>: <DSLField Query::hero>",
    ):
        ds.Human.name.select(ds.Query.hero)


def test_invalid_arg(ds):
    with pytest.raises(
        KeyError, match="Argument invalid_arg does not exist in Field: Character."
    ):
        ds.Query.hero.args(invalid_arg=5).select(ds.Character.name)


def test_multiple_root_fields(ds, client):
    query = dsl_gql(
        DSLQuery(
            ds.Query.hero.select(ds.Character.name),
            ds.Query.hero(episode=5)
            .alias("hero_of_episode_5")
            .select(ds.Character.name),
        )
    )
    result = client.execute(query)
    expected = {
        "hero": {"name": "R2-D2"},
        "hero_of_episode_5": {"name": "Luke Skywalker"},
    }
    assert result == expected
    assert node_tree(query.document) == node_tree(
        gql(print_ast(query.document)).document
    )


def test_root_fields_aliased(ds, client):
    query = dsl_gql(
        DSLQuery(
            ds.Query.hero.select(ds.Character.name),
            hero_of_episode_5=ds.Query.hero(episode=5).select(ds.Character.name),
        )
    )
    result = client.execute(query)
    expected = {
        "hero": {"name": "R2-D2"},
        "hero_of_episode_5": {"name": "Luke Skywalker"},
    }
    assert result == expected
    assert node_tree(query.document) == node_tree(
        gql(print_ast(query.document)).document
    )


def test_operation_name(ds):
    query = dsl_gql(
        GetHeroName=DSLQuery(
            ds.Query.hero.select(ds.Character.name),
        )
    )

    assert (
        print_ast(query.document)
        == """query GetHeroName {
  hero {
    name
  }
}"""
    )

    assert node_tree(query.document) == node_tree(
        gql(print_ast(query.document)).document
    )


def test_multiple_operations(ds):
    query = dsl_gql(
        GetHeroName=DSLQuery(ds.Query.hero.select(ds.Character.name)),
        CreateReviewMutation=DSLMutation(
            ds.Mutation.createReview.args(
                episode=6, review={"stars": 5, "commentary": "This is a great movie!"}
            ).select(ds.Review.stars, ds.Review.commentary)
        ),
    )

    assert (
        strip_braces_spaces(print_ast(query.document))
        == """query GetHeroName {
  hero {
    name
  }
}

mutation CreateReviewMutation {
  createReview(
    episode: JEDI
    review: {stars: 5, commentary: "This is a great movie!"}
  ) {
    stars
    commentary
  }
}"""
    )

    assert node_tree(query.document) == node_tree(
        gql(print_ast(query.document)).document
    )


def test_inline_fragments(ds):
    query = """hero(episode: JEDI) {
  name
  ... on Droid {
    primaryFunction
  }
  ... on Human {
    homePlanet
  }
}"""
    query_dsl = ds.Query.hero.args(episode=6).select(
        ds.Character.name,
        DSLInlineFragment().on(ds.Droid).select(ds.Droid.primaryFunction),
        DSLInlineFragment().on(ds.Human).select(ds.Human.homePlanet),
    )
    assert query == str(query_dsl)


def test_inline_fragments_nested(ds):
    query = """hero(episode: JEDI) {
  name
  ... on Human {
    ... on Human {
      homePlanet
    }
  }
}"""
    query_dsl = ds.Query.hero.args(episode=6).select(
        ds.Character.name,
        DSLInlineFragment()
        .on(ds.Human)
        .select(DSLInlineFragment().on(ds.Human).select(ds.Human.homePlanet)),
    )
    assert query == str(query_dsl)


def test_fragments_repr(ds):

    assert repr(DSLInlineFragment()) == "<DSLInlineFragment>"
    assert repr(DSLInlineFragment().on(ds.Droid)) == "<DSLInlineFragment on Droid>"
    assert repr(DSLFragment("fragment_1")) == "<DSLFragment fragment_1>"
    assert repr(DSLFragment("fragment_2").on(ds.Droid)) == "<DSLFragment fragment_2>"


def test_fragments(ds):
    query = """fragment NameAndAppearances on Character {
  name
  appearsIn
}

{
  hero {
    ...NameAndAppearances
  }
}"""

    name_and_appearances = (
        DSLFragment("NameAndAppearances")
        .on(ds.Character)
        .select(ds.Character.name, ds.Character.appearsIn)
    )

    query_dsl = DSLQuery(ds.Query.hero.select(name_and_appearances))

    request = dsl_gql(name_and_appearances, query_dsl)

    document = request.document

    print(print_ast(document))

    assert query == print_ast(document)
    assert node_tree(document) == node_tree(gql(print_ast(document)).document)


def test_fragment_without_type_condition_error(ds):

    # We create a fragment without using the .on(type_condition) method
    name_and_appearances = DSLFragment("NameAndAppearances")

    # If we try to use this fragment, gql generates an error
    with pytest.raises(
        AttributeError,
        match=r"Missing type condition. Please use .on\(type_condition\) method",
    ):
        dsl_gql(name_and_appearances)

    with pytest.raises(
        AttributeError,
        match=r"Missing type condition. Please use .on\(type_condition\) method",
    ):
        DSLFragment("NameAndAppearances").select(
            ds.Character.name, ds.Character.appearsIn
        )


def test_inline_fragment_in_dsl_gql(ds):

    inline_fragment = DSLInlineFragment()

    query = DSLQuery()

    with pytest.raises(
        GraphQLError,
        match=r"Invalid field for <DSLQuery>: <DSLInlineFragment>",
    ):
        query.select(inline_fragment)


def test_fragment_with_name_changed(ds):

    fragment = DSLFragment("ABC")

    assert str(fragment) == "...ABC"

    fragment.name = "DEF"

    assert str(fragment) == "...DEF"


def test_fragment_select_field_not_in_fragment(ds):

    fragment = DSLFragment("test").on(ds.Character)

    with pytest.raises(
        GraphQLError,
        match="Invalid field for <DSLFragment test>: <DSLField Droid::primaryFunction>",
    ):
        fragment.select(ds.Droid.primaryFunction)


def test_dsl_nested_query_with_fragment(ds):
    query = """fragment NameAndAppearances on Character {
  name
  appearsIn
}

query NestedQueryWithFragment {
  hero {
    ...NameAndAppearances
    friends {
      ...NameAndAppearances
      friends {
        ...NameAndAppearances
      }
    }
  }
}"""

    name_and_appearances = (
        DSLFragment("NameAndAppearances")
        .on(ds.Character)
        .select(ds.Character.name, ds.Character.appearsIn)
    )

    query_dsl = DSLQuery(
        ds.Query.hero.select(
            name_and_appearances,
            ds.Character.friends.select(
                name_and_appearances, ds.Character.friends.select(name_and_appearances)
            ),
        )
    )

    request = dsl_gql(name_and_appearances, NestedQueryWithFragment=query_dsl)

    document = request.document

    print(print_ast(document))

    assert query == print_ast(document)
    assert node_tree(document) == node_tree(gql(print_ast(document)).document)

    # Same thing, but incrementaly

    name_and_appearances = DSLFragment("NameAndAppearances")
    name_and_appearances.on(ds.Character)
    name_and_appearances.select(ds.Character.name)
    name_and_appearances.select(ds.Character.appearsIn)

    level_2 = ds.Character.friends
    level_2.select(name_and_appearances)
    level_1 = ds.Character.friends
    level_1.select(name_and_appearances)
    level_1.select(level_2)

    hero = ds.Query.hero
    hero.select(name_and_appearances)
    hero.select(level_1)

    query_dsl = DSLQuery(hero)

    request = dsl_gql(name_and_appearances, NestedQueryWithFragment=query_dsl)

    document = request.document

    print(print_ast(document))

    assert query == print_ast(document)
    assert node_tree(document) == node_tree(gql(print_ast(document)).document)


def test_dsl_query_all_fields_should_be_instances_of_DSLField():
    with pytest.raises(
        TypeError,
        match="Fields should be instances of DSLSelectable. Received: <class 'str'>",
    ):
        DSLQuery("I am a string")  # type: ignore


def test_dsl_query_all_fields_should_correspond_to_the_root_type(ds):
    with pytest.raises(GraphQLError) as excinfo:
        DSLQuery(ds.Character.name)

    assert ("Invalid field for <DSLQuery>: <DSLField Character::name>") in str(
        excinfo.value
    )


def test_dsl_root_type_not_default():

    schema_str = """
schema {
  query: QueryNotDefault
}

type QueryNotDefault {
  version: String
}
"""

    type_def_ast = parse(schema_str)
    schema = build_ast_schema(type_def_ast)

    ds = DSLSchema(schema)

    query = dsl_gql(DSLQuery(ds.QueryNotDefault.version))

    expected_query = """
{
  version
}
"""
    assert print_ast(query.document) == expected_query.strip()

    with pytest.raises(GraphQLError) as excinfo:
        DSLSubscription(ds.QueryNotDefault.version)

    assert (
        "Invalid field for <DSLSubscription>: <DSLField QueryNotDefault::version>"
    ) in str(excinfo.value)

    assert node_tree(query.document) == node_tree(
        gql(print_ast(query.document)).document
    )


def test_dsl_gql_all_arguments_should_be_operations_or_fragments():
    with pytest.raises(
        TypeError, match="Operations should be instances of DSLExecutable "
    ):
        dsl_gql("I am a string")  # type: ignore


def test_DSLSchema_requires_a_schema(client):
    with pytest.raises(TypeError, match="DSLSchema needs a schema as parameter"):
        DSLSchema(client)


def test_invalid_type(ds):
    with pytest.raises(
        AttributeError, match="Type 'invalid_type' not found in the schema!"
    ):
        ds.invalid_type


def test_invalid_type_union():
    schema_str = """
    type FloatValue {
        floatValue: Float!
    }

    type IntValue {
        intValue: Int!
    }

    union Value = FloatValue | IntValue

    type Entry {
        name: String!
        value: Value
    }

    type Query {
        values: [Entry!]!
    }
    """

    schema = build_ast_schema(parse(schema_str))
    ds = DSLSchema(schema)

    with pytest.raises(
        AttributeError,
        match=(
            "Type \"Value \\(<GraphQLUnionType 'Value'>\\)\" is not valid as an "
            "attribute of DSLSchema. Only Object types or Interface types are accepted."
        ),
    ):
        ds.Value


def test_hero_name_query_with_typename(ds):
    query = """
hero {
  name
  __typename
}
    """.strip()
    query_dsl = ds.Query.hero.select(ds.Character.name, DSLMetaField("__typename"))
    assert query == str(query_dsl)


def test_type_hero_query(ds):
    query = """{
  __type(name: "Hero") {
    kind
    name
    ofType {
      kind
      name
    }
  }
}"""

    type_hero = DSLMetaField("__type")(name="Hero")
    type_hero.select(
        ds.__Type.kind,
        ds.__Type.name,
        ds.__Type.ofType.select(ds.__Type.kind, ds.__Type.name),
    )
    query_dsl = DSLQuery(type_hero)

    assert query == str(print_ast(dsl_gql(query_dsl).document)).strip()


def test_invalid_meta_field_selection(ds):

    DSLQuery(DSLMetaField("__typename"))
    DSLQuery(DSLMetaField("__schema"))
    DSLQuery(DSLMetaField("__type"))

    metafield = DSLMetaField("__typename")
    assert metafield.name == "__typename"

    with pytest.raises(GraphQLError):
        DSLMetaField("__invalid_meta_field")

    DSLMutation(DSLMetaField("__typename"))

    with pytest.raises(GraphQLError):
        DSLMutation(DSLMetaField("__schema"))

    with pytest.raises(GraphQLError):
        DSLMutation(DSLMetaField("__type"))

    with pytest.raises(GraphQLError):
        DSLSubscription(DSLMetaField("__typename"))

    with pytest.raises(GraphQLError):
        DSLSubscription(DSLMetaField("__schema"))

    with pytest.raises(GraphQLError):
        DSLSubscription(DSLMetaField("__type"))

    fragment = DSLFragment("blah")

    with pytest.raises(AttributeError):
        fragment.select(DSLMetaField("__typename"))

    fragment.on(ds.Character)

    fragment.select(DSLMetaField("__typename"))

    with pytest.raises(GraphQLError):
        fragment.select(DSLMetaField("__schema"))

    with pytest.raises(GraphQLError):
        fragment.select(DSLMetaField("__type"))

    ds.Query.hero.select(DSLMetaField("__typename"))

    with pytest.raises(GraphQLError):
        ds.Query.hero.select(DSLMetaField("__schema"))

    with pytest.raises(GraphQLError):
        ds.Query.hero.select(DSLMetaField("__type"))


@pytest.mark.parametrize("option", [True, False])
def test_get_introspection_query_ast(option):

    introspection_query = get_introspection_query(
        descriptions=option,
        specified_by_url=option,
        directive_is_repeatable=option,
        schema_description=option,
        input_value_deprecation=option,
    )
    dsl_introspection_query = get_introspection_query_ast(
        descriptions=option,
        specified_by_url=option,
        directive_is_repeatable=option,
        schema_description=option,
        input_value_deprecation=option,
    )

    try:
        assert print_ast(gql(introspection_query).document) == print_ast(
            dsl_introspection_query
        )
        assert node_tree(dsl_introspection_query) == node_tree(
            gql(print_ast(dsl_introspection_query)).document
        )
    except AssertionError:

        # From graphql-core version 3.3.0a7, there is two more type recursion levels
        dsl_introspection_query = get_introspection_query_ast(
            descriptions=option,
            specified_by_url=option,
            directive_is_repeatable=option,
            schema_description=option,
            input_value_deprecation=option,
            type_recursion_level=9,
        )
        assert print_ast(gql(introspection_query).document) == print_ast(
            dsl_introspection_query
        )
        assert node_tree(dsl_introspection_query) == node_tree(
            gql(print_ast(dsl_introspection_query)).document
        )


def test_typename_aliased(ds):
    query = """
hero {
  name
  typenameField: __typename
}
""".strip()

    query_dsl = ds.Query.hero.select(
        ds.Character.name, typenameField=DSLMetaField("__typename")
    )
    assert query == str(query_dsl)

    query_dsl = ds.Query.hero.select(
        ds.Character.name, DSLMetaField("__typename").alias("typenameField")
    )
    assert query == str(query_dsl)


def test_node_tree_with_loc(ds):
    query = """query GetHeroName {
  hero {
    name
  }
}""".strip()

    document = gql(query).document

    node_tree_result = """
DocumentNode
  definitions:
    OperationDefinitionNode
      directives:
        empty tuple
      loc:
        Location
          <Location 0:43>
      name:
        NameNode
          loc:
            Location
              <Location 6:17>
          value:
            'GetHeroName'
      operation:
        <OperationType.QUERY: 'query'>
      selection_set:
        SelectionSetNode
          loc:
            Location
              <Location 18:43>
          selections:
            FieldNode
              alias:
                None
              arguments:
                empty tuple
              directives:
                empty tuple
              loc:
                Location
                  <Location 22:41>
              name:
                NameNode
                  loc:
                    Location
                      <Location 22:26>
                  value:
                    'hero'
              nullability_assertion:
                None
              selection_set:
                SelectionSetNode
                  loc:
                    Location
                      <Location 27:41>
                  selections:
                    FieldNode
                      alias:
                        None
                      arguments:
                        empty tuple
                      directives:
                        empty tuple
                      loc:
                        Location
                          <Location 33:37>
                      name:
                        NameNode
                          loc:
                            Location
                              <Location 33:37>
                          value:
                            'name'
                      nullability_assertion:
                        None
                      selection_set:
                        None
      variable_definitions:
        empty tuple
  loc:
    Location
      <Location 0:43>
""".strip()

    node_tree_result_stable = """
DocumentNode
  definitions:
    OperationDefinitionNode
      directives:
        empty tuple
      loc:
        Location
          <Location 0:43>
      name:
        NameNode
          loc:
            Location
              <Location 6:17>
          value:
            'GetHeroName'
      operation:
        <OperationType.QUERY: 'query'>
      selection_set:
        SelectionSetNode
          loc:
            Location
              <Location 18:43>
          selections:
            FieldNode
              alias:
                None
              arguments:
                empty tuple
              directives:
                empty tuple
              loc:
                Location
                  <Location 22:41>
              name:
                NameNode
                  loc:
                    Location
                      <Location 22:26>
                  value:
                    'hero'
              selection_set:
                SelectionSetNode
                  loc:
                    Location
                      <Location 27:41>
                  selections:
                    FieldNode
                      alias:
                        None
                      arguments:
                        empty tuple
                      directives:
                        empty tuple
                      loc:
                        Location
                          <Location 33:37>
                      name:
                        NameNode
                          loc:
                            Location
                              <Location 33:37>
                          value:
                            'name'
                      selection_set:
                        None
      variable_definitions:
        empty tuple
  loc:
    Location
      <Location 0:43>
""".strip()

    print(node_tree(document, ignore_loc=False))

    try:
        assert node_tree(document, ignore_loc=False) == node_tree_result
    except AssertionError:
        # graphql-core version 3.2.3
        assert node_tree(document, ignore_loc=False) == node_tree_result_stable


def test_legacy_fragment_with_variables(ds):
    var = DSLVariableDefinitions()

    hero_fragment = (
        DSLFragment("heroFragment")
        .on(ds.Query)
        .select(
            ds.Query.hero.args(episode=var.episode).select(ds.Character.name),
        )
    )

    print(hero_fragment)

    hero_fragment.variable_definitions = var

    query = dsl_gql(hero_fragment)

    expected = """
fragment heroFragment($episode: Episode) on Query {
  hero(episode: $episode) {
    name
  }
}
""".strip()
    assert print_ast(query.document) == expected



import pytest


@pytest.fixture
def fake_signer_factory(fake_request_factory):
    def _fake_signer_factory(request=None):
        if not request:
            request = fake_request_factory()
        return FakeSigner(request=request)

    yield _fake_signer_factory


class FakeSigner:
    def __init__(self, request=None):
        self.request = request

    def add_auth(self, request):
        """
        A fake for getting a request object that
        :return:
        """
        request.headers = {"FakeAuthorization": "a", "FakeTime": "today"}

    def get_headers(self):
        self.add_auth(self.request)
        return self.request.headers



import asyncio
import io
import json
import logging
from ssl import SSLContext
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

import aiohttp
from aiohttp.client_exceptions import ClientResponseError
from aiohttp.client_reqrep import Fingerprint
from aiohttp.helpers import BasicAuth
from aiohttp.typedefs import LooseCookies, LooseHeaders
from graphql import ExecutionResult
from multidict import CIMultiDictProxy

from ..graphql_request import GraphQLRequest
from .appsync_auth import AppSyncAuthentication
from .async_transport import AsyncTransport
from .common.aiohttp_closed_event import create_aiohttp_closed_event
from .common.batch import get_batch_execution_result_list
from .exceptions import (
    TransportAlreadyConnected,
    TransportClosed,
    TransportConnectionFailed,
    TransportError,
    TransportProtocolError,
    TransportServerError,
)
from .file_upload import FileVar, close_files, extract_files, open_files

log = logging.getLogger(__name__)


class AIOHTTPTransport(AsyncTransport):
    """:ref:`Async Transport <async_transports>` to execute GraphQL queries
    on remote servers with an HTTP connection.

    This transport use the aiohttp library with asyncio.
    """

    file_classes: Tuple[Type[Any], ...] = (
        io.IOBase,
        aiohttp.StreamReader,
        AsyncGenerator,
    )

    def __init__(
        self,
        url: str,
        headers: Optional[LooseHeaders] = None,
        cookies: Optional[LooseCookies] = None,
        auth: Optional[Union[BasicAuth, "AppSyncAuthentication"]] = None,
        ssl: Union[SSLContext, bool, Fingerprint] = True,
        timeout: Optional[int] = None,
        ssl_close_timeout: Optional[Union[int, float]] = 10,
        json_serialize: Callable = json.dumps,
        json_deserialize: Callable = json.loads,
        client_session_args: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Initialize the transport with the given aiohttp parameters.

        :param url: The GraphQL server URL. Example: 'https://server.com:PORT/path'.
        :param headers: Dict of HTTP Headers.
        :param cookies: Dict of HTTP cookies.
        :param auth: BasicAuth object to enable Basic HTTP auth if needed
                     Or Appsync Authentication class
        :param ssl: ssl_context of the connection.
                    Use ssl=False to not verify ssl certificates.
        :param ssl_close_timeout: Timeout in seconds to wait for the ssl connection
                                  to close properly
        :param json_serialize: Json serializer callable.
                By default json.dumps() function
        :param json_deserialize: Json deserializer callable.
                By default json.loads() function
        :param client_session_args: Dict of extra args passed to
                `aiohttp.ClientSession`_

        .. _aiohttp.ClientSession:
          https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.ClientSession
        """
        self.url: str = url
        self.headers: Optional[LooseHeaders] = headers
        self.cookies: Optional[LooseCookies] = cookies
        self.auth: Optional[Union[BasicAuth, "AppSyncAuthentication"]] = auth
        self.ssl: Union[SSLContext, bool, Fingerprint] = ssl
        self.timeout: Optional[int] = timeout
        self.ssl_close_timeout: Optional[Union[int, float]] = ssl_close_timeout
        self.client_session_args = client_session_args
        self.session: Optional[aiohttp.ClientSession] = None
        self.response_headers: Optional[CIMultiDictProxy[str]]
        self.json_serialize: Callable = json_serialize
        self.json_deserialize: Callable = json_deserialize

    async def connect(self) -> None:
        """Coroutine which will create an aiohttp ClientSession() as self.session.

        Don't call this coroutine directly on the transport, instead use
        :code:`async with` on the client and this coroutine will be executed
        to create the session.

        Should be cleaned with a call to the close coroutine.
        """

        if self.session is None:

            client_session_args: Dict[str, Any] = {
                "cookies": self.cookies,
                "headers": self.headers,
                "auth": (
                    None if isinstance(self.auth, AppSyncAuthentication) else self.auth
                ),
                "json_serialize": self.json_serialize,
            }

            if self.timeout is not None:
                client_session_args["timeout"] = aiohttp.ClientTimeout(
                    total=self.timeout
                )

            # Adding custom parameters passed from init
            if self.client_session_args:
                client_session_args.update(self.client_session_args)

            log.debug("Connecting transport")

            self.session = aiohttp.ClientSession(**client_session_args)

        else:
            raise TransportAlreadyConnected("Transport is already connected")

    async def close(self) -> None:
        """Coroutine which will close the aiohttp session.

        Don't call this coroutine directly on the transport, instead use
        :code:`async with` on the client and this coroutine will be executed
        when you exit the async context manager.
        """
        if self.session is not None:

            log.debug("Closing transport")

            if (
                self.client_session_args
                and self.client_session_args.get("connector_owner") is False
            ):

                log.debug("connector_owner is False -> not closing connector")

            else:
                closed_event = create_aiohttp_closed_event(self.session)
                await self.session.close()
                try:
                    await asyncio.wait_for(closed_event.wait(), self.ssl_close_timeout)
                except asyncio.TimeoutError:
                    pass

        self.session = None

    def _prepare_request(
        self,
        request: Union[GraphQLRequest, List[GraphQLRequest]],
        extra_args: Optional[Dict[str, Any]] = None,
        upload_files: bool = False,
    ) -> Dict[str, Any]:

        payload: Dict | List
        if isinstance(request, GraphQLRequest):
            payload = request.payload
        else:
            payload = [req.payload for req in request]

        if upload_files:
            assert isinstance(payload, Dict)
            assert isinstance(request, GraphQLRequest)
            post_args = self._prepare_file_uploads(request, payload)
        else:
            post_args = {"json": payload}

        # Log the payload
        if log.isEnabledFor(logging.DEBUG):
            log.debug(">>> %s", self.json_serialize(payload))

        # Pass post_args to aiohttp post method
        if extra_args:
            post_args.update(extra_args)

        # Add headers for AppSync if requested
        if isinstance(self.auth, AppSyncAuthentication):
            post_args["headers"] = self.auth.get_headers(
                self.json_serialize(payload),
                {"content-type": "application/json"},
            )

        return post_args

    def _prepare_file_uploads(
        self, request: GraphQLRequest, payload: Dict[str, Any]
    ) -> Dict[str, Any]:

        # If the upload_files flag is set, then we need variable_values
        variable_values = request.variable_values
        assert variable_values is not None

        # If we upload files, we will extract the files present in the
        # variable_values dict and replace them by null values
        nulled_variable_values, files = extract_files(
            variables=variable_values,
            file_classes=self.file_classes,
        )

        # Opening the files using the FileVar parameters
        open_files(list(files.values()), transport_supports_streaming=True)
        self.files = files

        # Save the nulled variable values in the payload
        payload["variables"] = nulled_variable_values

        # Prepare aiohttp to send multipart-encoded data
        data = aiohttp.FormData()

        # Generate the file map
        # path is nested in a list because the spec allows multiple pointers
        # to the same file. But we don't support that.
        # Will generate something like {"0": ["variables.file"]}
        file_map = {str(i): [path] for i, path in enumerate(files)}

        # Enumerate the file streams
        # Will generate something like {'0': FileVar object}
        file_vars = {str(i): files[path] for i, path in enumerate(files)}

        # Add the payload to the operations field
        operations_str = self.json_serialize(payload)
        log.debug("operations %s", operations_str)
        data.add_field("operations", operations_str, content_type="application/json")

        # Add the file map field
        file_map_str = self.json_serialize(file_map)
        log.debug("file_map %s", file_map_str)
        data.add_field("map", file_map_str, content_type="application/json")

        for k, file_var in file_vars.items():
            assert isinstance(file_var, FileVar)

            data.add_field(
                k,
                file_var.f,
                filename=file_var.filename,
                content_type=file_var.content_type,
            )

        post_args: Dict[str, Any] = {"data": data}

        return post_args

    @staticmethod
    def _raise_transport_server_error_if_status_more_than_400(
        resp: aiohttp.ClientResponse,
    ) -> None:
        # If the status is >400,
        # then we need to raise a TransportServerError
        try:
            # Raise ClientResponseError if response status is 400 or higher
            resp.raise_for_status()
        except ClientResponseError as e:
            raise TransportServerError(str(e), e.status) from e

    @classmethod
    async def _raise_response_error(
        cls,
        resp: aiohttp.ClientResponse,
        reason: str,
    ) -> None:
        # We raise a TransportServerError if status code is 400 or higher
        # We raise a TransportProtocolError in the other cases

        cls._raise_transport_server_error_if_status_more_than_400(resp)

        result_text = await resp.text()
        raise TransportProtocolError(
            f"Server did not return a valid GraphQL result: "
            f"{reason}: "
            f"{result_text}"
        )

    async def _get_json_result(self, response: aiohttp.ClientResponse) -> Any:

        # Saving latest response headers in the transport
        self.response_headers = response.headers

        try:
            result = await response.json(loads=self.json_deserialize, content_type=None)

            if log.isEnabledFor(logging.DEBUG):
                result_text = await response.text()
                log.debug("<<< %s", result_text)

        except Exception:
            await self._raise_response_error(response, "Not a JSON answer")

        if result is None:
            await self._raise_response_error(response, "Not a JSON answer")

        return result

    async def _prepare_result(
        self, response: aiohttp.ClientResponse
    ) -> ExecutionResult:

        result = await self._get_json_result(response)

        if "errors" not in result and "data" not in result:
            await self._raise_response_error(
                response, 'No "data" or "errors" keys in answer'
            )

        return ExecutionResult(
            errors=result.get("errors"),
            data=result.get("data"),
            extensions=result.get("extensions"),
        )

    async def _prepare_batch_result(
        self,
        reqs: List[GraphQLRequest],
        response: aiohttp.ClientResponse,
    ) -> List[ExecutionResult]:

        answers = await self._get_json_result(response)

        try:
            return get_batch_execution_result_list(reqs, answers)
        except TransportProtocolError:
            # Raise a TransportServerError if status > 400
            self._raise_transport_server_error_if_status_more_than_400(response)
            # In other cases, raise a TransportProtocolError
            raise

    async def execute(
        self,
        request: GraphQLRequest,
        *,
        extra_args: Optional[Dict[str, Any]] = None,
        upload_files: bool = False,
    ) -> ExecutionResult:
        """Execute the provided request against the configured remote server
        using the current session.
        This uses the aiohttp library to perform a HTTP POST request asynchronously
        to the remote server.

        Don't call this coroutine directly on the transport, instead use
        :code:`execute` on a client or a session.

        :param request: GraphQL request as a
                        :class:`GraphQLRequest <gql.GraphQLRequest>` object.
        :param extra_args: additional arguments to send to the aiohttp post method
        :param upload_files: Set to True if you want to put files in the variable values
        :returns: an ExecutionResult object.
        """

        if self.session is None:
            raise TransportClosed("Transport is not connected")

        post_args = self._prepare_request(
            request,
            extra_args,
            upload_files,
        )

        try:
            async with self.session.post(self.url, ssl=self.ssl, **post_args) as resp:
                return await self._prepare_result(resp)
        except TransportError:
            raise
        except Exception as e:
            raise TransportConnectionFailed(str(e)) from e
        finally:
            if upload_files:
                close_files(list(self.files.values()))

    async def execute_batch(
        self,
        reqs: List[GraphQLRequest],
        extra_args: Optional[Dict[str, Any]] = None,
    ) -> List[ExecutionResult]:
        """Execute multiple GraphQL requests in a batch.

        Don't call this coroutine directly on the transport, instead use
        :code:`execute_batch` on a client or a session.

        :param reqs: GraphQL requests as a list of GraphQLRequest objects.
        :param extra_args: additional arguments to send to the aiohttp post method
        :return: A list of results of execution.
            For every result `data` is the result of executing the query,
            `errors` is null if no errors occurred, and is a non-empty array
            if an error occurred.
        """

        if self.session is None:
            raise TransportClosed("Transport is not connected")

        post_args = self._prepare_request(
            reqs,
            extra_args,
        )

        try:
            async with self.session.post(self.url, ssl=self.ssl, **post_args) as resp:
                return await self._prepare_batch_result(reqs, resp)
        except TransportError:
            raise
        except Exception as e:
            raise TransportConnectionFailed(str(e)) from e

    def subscribe(
        self,
        request: GraphQLRequest,
    ) -> AsyncGenerator[ExecutionResult, None]:
        """Subscribe is not supported on HTTP.

        :meta private:
        """
        raise NotImplementedError(" The HTTP transport does not support subscriptions")



import pytest

from gql import Client, gql


@pytest.mark.aiohttp
@pytest.mark.online
@pytest.mark.asyncio
@pytest.mark.parametrize("fetch_schema_from_transport", [True, False])
async def test_async_client_async_transport(fetch_schema_from_transport):

    from gql.transport.aiohttp import AIOHTTPTransport

    # Create https url
    url = "https://countries.trevorblades.com/graphql"

    # Get async transport
    transport = AIOHTTPTransport(url=url)

    # Instantiate client
    async with Client(
        transport=transport,
        fetch_schema_from_transport=fetch_schema_from_transport,
    ) as session:

        query = gql(
            """
            query getContinents {
              continents {
                code
                name
              }
            }
        """
        )

        # Execute query
        result = await session.execute(query)

        continents = result["continents"]

        africa = continents[0]

        assert africa["code"] == "AF"

        if fetch_schema_from_transport:
            assert session.client.schema is not None


@pytest.mark.requests
@pytest.mark.online
@pytest.mark.asyncio
@pytest.mark.parametrize("fetch_schema_from_transport", [True, False])
async def test_async_client_sync_transport(fetch_schema_from_transport):

    from gql.transport.requests import RequestsHTTPTransport

    url = "http://countries.trevorblades.com/graphql"

    # Get sync transport
    transport = RequestsHTTPTransport(url=url, use_json=True)

    # Impossible to use a sync transport asynchronously
    with pytest.raises(AssertionError):
        async with Client(
            transport=transport,
            fetch_schema_from_transport=fetch_schema_from_transport,
        ):
            pass

    transport.close()


@pytest.mark.aiohttp
@pytest.mark.online
@pytest.mark.parametrize("fetch_schema_from_transport", [True, False])
def test_sync_client_async_transport(fetch_schema_from_transport):

    from gql.transport.aiohttp import AIOHTTPTransport

    # Create https url
    url = "https://countries.trevorblades.com/graphql"

    # Get async transport
    transport = AIOHTTPTransport(url=url)

    # Instanciate client
    client = Client(
        transport=transport,
        fetch_schema_from_transport=fetch_schema_from_transport,
    )

    query = gql(
        """
        query getContinents {
          continents {
            code
            name
          }
        }
    """
    )

    # Execute query synchronously
    result = client.execute(query)

    continents = result["continents"]

    africa = continents[0]

    assert africa["code"] == "AF"

    if fetch_schema_from_transport:
        assert client.schema is not None


@pytest.mark.requests
@pytest.mark.online
@pytest.mark.parametrize("fetch_schema_from_transport", [True, False])
def test_sync_client_sync_transport(fetch_schema_from_transport):

    from gql.transport.requests import RequestsHTTPTransport

    # Create https url
    url = "https://countries.trevorblades.com/graphql"

    # Get sync transport
    transport = RequestsHTTPTransport(url=url, use_json=True)

    # Instanciate client
    client = Client(
        transport=transport,
        fetch_schema_from_transport=fetch_schema_from_transport,
    )

    query = gql(
        """
        query getContinents {
          continents {
            code
            name
          }
        }
    """
    )

    # Execute query synchronously
    result = client.execute(query)

    continents = result["continents"]

    africa = continents[0]

    assert africa["code"] == "AF"

    if fetch_schema_from_transport:
        assert client.schema is not None



import logging

import pytest

from gql import __version__
from gql.cli import (
    get_execute_args,
    get_introspection_args,
    get_parser,
    get_transport,
    get_transport_args,
    main,
)


@pytest.fixture
def parser():
    return get_parser()


def test_cli_parser(parser):

    # Simple call with https server
    # gql-cli https://your_server.com
    args = parser.parse_args(["https://your_server.com"])

    assert args.server == "https://your_server.com"
    assert args.headers is None
    assert args.loglevel is None
    assert args.operation_name is None
    assert args.variables is None

    # Call with variable values parameters
    # gql-cli https://your_server.com --variables KEY1:value1 KEY2:value2
    args = parser.parse_args(
        ["https://your_server.com", "--variables", "KEY1:value1", "KEY2:value2"]
    )

    assert args.server == "https://your_server.com"
    assert args.variables == ["KEY1:value1", "KEY2:value2"]

    # Call with headers values parameters
    # gql-cli https://your_server.com --headers HEADER1:value1 HEADER2:value2
    args = parser.parse_args(
        ["https://your_server.com", "--headers", "HEADER1:value1", "HEADER2:value2"]
    )

    assert args.server == "https://your_server.com"
    assert args.headers == ["HEADER1:value1", "HEADER2:value2"]

    # Call with header value with a space in value
    # gql-cli https://your_server.com --headers Authorization:"Bearer blahblah"
    args = parser.parse_args(
        ["https://your_server.com", "--headers", "Authorization:Bearer blahblah"]
    )

    assert args.server == "https://your_server.com"
    assert args.headers == ["Authorization:Bearer blahblah"]

    # Check loglevel flags
    # gql-cli https://your_server.com --debug
    args = parser.parse_args(["https://your_server.com", "--debug"])
    assert args.loglevel == logging.DEBUG

    # gql-cli https://your_server.com --verbose
    args = parser.parse_args(["https://your_server.com", "--verbose"])

    assert args.loglevel == logging.INFO

    # Check operation_name
    # gql-cli https://your_server.com --operation-name my_operation
    args = parser.parse_args(
        ["https://your_server.com", "--operation-name", "my_operation"]
    )
    assert args.operation_name == "my_operation"

    # Check execute_timeout
    # gql-cli https://your_server.com --execute-timeout 1
    args = parser.parse_args(["https://your_server.com", "--execute-timeout", "1"])
    assert args.execute_timeout == 1

    # gql-cli https://your_server.com --execute-timeout=none
    args = parser.parse_args(["https://your_server.com", "--execute-timeout", "none"])
    assert args.execute_timeout is None

    # gql-cli https://your_server.com --execute-timeout=-1
    with pytest.raises(SystemExit):
        args = parser.parse_args(["https://your_server.com", "--execute-timeout", "-1"])

    # gql-cli https://your_server.com --execute-timeout=invalid
    with pytest.raises(SystemExit):
        args = parser.parse_args(
            ["https://your_server.com", "--execute-timeout", "invalid"]
        )


def test_cli_parse_headers(parser):

    args = parser.parse_args(
        [
            "https://your_server.com",
            "--headers",
            "Token1:1234",
            "Token2:5678",
            "TokenWithSpace:abc def",
            "TokenWithColon:abc:def",
        ]
    )

    transport_args = get_transport_args(args)

    expected_headers = {
        "Token1": "1234",
        "Token2": "5678",
        "TokenWithSpace": "abc def",
        "TokenWithColon": "abc:def",
    }

    assert transport_args == {"headers": expected_headers}


def test_cli_parse_headers_invalid_header(parser):

    args = parser.parse_args(
        ["https://your_server.com", "--headers", "TokenWithoutColon"]
    )

    with pytest.raises(ValueError):
        get_transport_args(args)


def test_cli_parse_operation_name(parser):

    args = parser.parse_args(["https://your_server.com", "--operation-name", "myop"])

    execute_args = get_execute_args(args)

    assert execute_args == {"operation_name": "myop"}


@pytest.mark.parametrize(
    "param",
    [
        {"args": ["key:abcdef"], "d": {"key": "abcdef"}},
        {"args": ['key:"abcdef"'], "d": {"key": "abcdef"}},
        {"args": ["key:1234"], "d": {"key": 1234}},
        {"args": ["key1:1234", "key2:5678"], "d": {"key1": 1234, "key2": 5678}},
        {"args": ["key1:null"], "d": {"key1": None}},
        {"args": ["key1:true"], "d": {"key1": True}},
        {"args": ["key1:false"], "d": {"key1": False}},
        {
            "args": ["key1:null", "key2:abcd", "key3:5"],
            "d": {"key1": None, "key2": "abcd", "key3": 5},
        },
    ],
)
def test_cli_parse_variable_value(parser, param):

    args = parser.parse_args(["https://your_server.com", "--variables", *param["args"]])

    execute_args = get_execute_args(args)

    expected_variable_values = param["d"]

    assert execute_args == {"variable_values": expected_variable_values}


@pytest.mark.parametrize("param", ["nocolon", 'key:"'])
def test_cli_parse_variable_value_invalid_param(parser, param):

    args = parser.parse_args(["https://your_server.com", "--variables", param])

    with pytest.raises(ValueError):
        get_execute_args(args)


@pytest.mark.aiohttp
@pytest.mark.parametrize(
    "url",
    ["http://your_server.com", "https://your_server.com"],
)
def test_cli_get_transport_aiohttp(parser, url):

    from gql.transport.aiohttp import AIOHTTPTransport

    args = parser.parse_args([url])

    transport = get_transport(args)

    assert isinstance(transport, AIOHTTPTransport)


@pytest.mark.httpx
@pytest.mark.parametrize(
    "url",
    ["http://your_server.com", "https://your_server.com"],
)
def test_cli_get_transport_httpx(parser, url):

    from gql.transport.httpx import HTTPXAsyncTransport

    args = parser.parse_args([url, "--transport", "httpx"])

    transport = get_transport(args)

    assert isinstance(transport, HTTPXAsyncTransport)


@pytest.mark.websockets
@pytest.mark.parametrize(
    "url",
    ["ws://your_server.com", "wss://your_server.com"],
)
def test_cli_get_transport_websockets(parser, url):

    from gql.transport.websockets import WebsocketsTransport

    args = parser.parse_args([url])

    transport = get_transport(args)

    assert isinstance(transport, WebsocketsTransport)


@pytest.mark.websockets
@pytest.mark.parametrize(
    "url",
    ["ws://your_server.com", "wss://your_server.com"],
)
def test_cli_get_transport_phoenix(parser, url):

    from gql.transport.phoenix_channel_websockets import (
        PhoenixChannelWebsocketsTransport,
    )

    args = parser.parse_args([url, "--transport", "phoenix"])

    transport = get_transport(args)

    assert isinstance(transport, PhoenixChannelWebsocketsTransport)


@pytest.mark.websockets
@pytest.mark.botocore
@pytest.mark.parametrize(
    "url",
    [
        "wss://XXXXXX.appsync-api.eu-west-3.amazonaws.com/graphql",
        "wss://noregion.amazonaws.com/graphql",
    ],
)
def test_cli_get_transport_appsync_websockets_iam(parser, url):

    args = parser.parse_args([url, "--transport", "appsync_websockets"])

    transport = get_transport(args)

    # In the tests, the AWS Appsync credentials are not set
    # So the transport is None
    assert transport is None


@pytest.mark.asyncio
@pytest.mark.websockets
@pytest.mark.botocore
@pytest.mark.parametrize(
    "url",
    ["wss://XXXXXX.appsync-api.eu-west-3.amazonaws.com/graphql"],
)
async def test_cli_main_appsync_websockets_iam(parser, url):

    args = parser.parse_args([url, "--transport", "appsync_websockets"])

    exit_code = await main(args)

    # In the tests, the AWS Appsync credentials are not set
    # So the transport is None and the main returns
    # an exit_code of 1
    assert exit_code == 1


@pytest.mark.websockets
@pytest.mark.parametrize(
    "url",
    ["wss://XXXXXX.appsync-api.eu-west-3.amazonaws.com/graphql"],
)
def test_cli_get_transport_appsync_websockets_api_key(parser, url):

    from gql.transport.appsync_auth import AppSyncApiKeyAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

    args = parser.parse_args(
        [url, "--transport", "appsync_websockets", "--api-key", "test-api-key"]
    )

    transport = get_transport(args)

    assert isinstance(transport, AppSyncWebsocketsTransport)
    assert isinstance(transport.auth, AppSyncApiKeyAuthentication)
    assert transport.auth.api_key == "test-api-key"


@pytest.mark.websockets
@pytest.mark.parametrize(
    "url",
    ["wss://XXXXXX.appsync-api.eu-west-3.amazonaws.com/graphql"],
)
def test_cli_get_transport_appsync_websockets_jwt(parser, url):

    from gql.transport.appsync_auth import AppSyncJWTAuthentication
    from gql.transport.appsync_websockets import AppSyncWebsocketsTransport

    args = parser.parse_args(
        [url, "--transport", "appsync_websockets", "--jwt", "test-jwt"]
    )

    transport = get_transport(args)

    assert isinstance(transport, AppSyncWebsocketsTransport)
    assert isinstance(transport.auth, AppSyncJWTAuthentication)
    assert transport.auth.jwt == "test-jwt"


@pytest.mark.aiohttp
@pytest.mark.botocore
@pytest.mark.parametrize(
    "url",
    ["https://XXXXXX.appsync-api.eu-west-3.amazonaws.com/graphql"],
)
def test_cli_get_transport_appsync_http_iam(parser, url):

    from gql.transport.aiohttp import AIOHTTPTransport

    args = parser.parse_args([url, "--transport", "appsync_http"])

    transport = get_transport(args)

    assert isinstance(transport, AIOHTTPTransport)


@pytest.mark.aiohttp
@pytest.mark.parametrize(
    "url",
    ["https://XXXXXX.appsync-api.eu-west-3.amazonaws.com/graphql"],
)
def test_cli_get_transport_appsync_http_api_key(parser, url):

    from gql.transport.aiohttp import AIOHTTPTransport
    from gql.transport.appsync_auth import AppSyncApiKeyAuthentication

    args = parser.parse_args(
        [url, "--transport", "appsync_http", "--api-key", "test-api-key"]
    )

    transport = get_transport(args)

    assert isinstance(transport, AIOHTTPTransport)
    assert isinstance(transport.auth, AppSyncApiKeyAuthentication)
    assert transport.auth.api_key == "test-api-key"


@pytest.mark.aiohttp
@pytest.mark.parametrize(
    "url",
    ["https://XXXXXX.appsync-api.eu-west-3.amazonaws.com/graphql"],
)
def test_cli_get_transport_appsync_http_jwt(parser, url):

    from gql.transport.aiohttp import AIOHTTPTransport
    from gql.transport.appsync_auth import AppSyncJWTAuthentication

    args = parser.parse_args([url, "--transport", "appsync_http", "--jwt", "test-jwt"])

    transport = get_transport(args)

    assert isinstance(transport, AIOHTTPTransport)
    assert isinstance(transport.auth, AppSyncJWTAuthentication)
    assert transport.auth.jwt == "test-jwt"


def test_cli_get_transport_no_protocol(parser):

    args = parser.parse_args(["your_server.com"])

    with pytest.raises(ValueError):
        get_transport(args)


@pytest.mark.script_launch_mode("subprocess")
def test_cli_ep_version(script_runner):
    ret = script_runner.run(["gql-cli", "--version"])

    assert ret.success

    assert ret.stdout == f"v{__version__}\n"
    assert ret.stderr == ""


def test_cli_parse_schema_download(parser):

    args = parser.parse_args(
        [
            "https://your_server.com",
            "--schema-download",
            "descriptions:false",
            "input_value_deprecation:true",
            "specified_by_url:True",
            "schema_description:true",
            "directive_is_repeatable:true",
            "--print-schema",
        ]
    )

    introspection_args = get_introspection_args(args)

    expected_args = {
        "descriptions": False,
        "input_value_deprecation": True,
        "specified_by_url": True,
        "schema_description": True,
        "directive_is_repeatable": True,
    }

    assert introspection_args == expected_args


@pytest.mark.parametrize(
    "invalid_args",
    [
        ["https://your_server.com", "--schema-download", "ArgWithoutColon"],
        ["https://your_server.com", "--schema-download", "blahblah:true"],
        ["https://your_server.com", "--schema-download", "descriptions:invalid_bool"],
    ],
)
def test_cli_parse_schema_download_invalid_arg(parser, invalid_args):
    args = parser.parse_args(invalid_args)

    with pytest.raises(ValueError):
        get_introspection_args(args)
